#!/usr/bin/env python
"""Summarize the mapping results of trimmed versus untrimmed reads.
This requires that both input bams be name-sorted.
"""
import argparse
from collections import defaultdict
from common import *
import csv
from glob import glob
import os

# There is the option to compute edit distance between the untrimmed and 
# trimmed reads. We did not use those metrics in the paper. If you want to 
# enable the edit distance calculation, you need to install the 'editdistance'
# pyton library.
#
# pip install editdistance
#
# # Note: if this doesn't work for you, you'll need to checkout the 
# editdistance repository and edit setup.py to enable the 'cythonize' command, 
# which will recompile the cython code for your local environment. So you would 
# run:
# 
# python setup.py build_ext -i && python setup.py install

nuc = ('A','C','G','T','N')

prog_header = ['prog', 'prog2', 'threads', 'dataset', 'qcut']

pair_fields = [
    ('prog', None),
    ('read_name', None),
    ('read_idx', None),
    ('failed', False),
    ('discarded', False),
    ('split', False),
    ('skipped', False),
    ('proper', False),
    ('valid', False)
]
read_fields = [
    ('mapped', False),
    ('quality', -1),
    ('score', -1),
    ('chrm', -1),
    ('start', -1),
    ('end', -1),
    ('clipped_front', -1),
    ('clipped_back', -1),
    ('trimmed_front', 0),
    ('trimmed_back', 0),
    ('in_region', -1)
]

def get_header():
    rf = [key for key, val in read_fields]
    r1 = ['read1_{}'.format(key) for key in rf]
    r2 = ['read2_{}'.format(key) for key in rf]
    return prog_header + [key for key, val in pair_fields[1:]] + r1 + r2

HEADER = get_header()

class TableRead(object):
    def __init__(self):
        for key, val in read_fields:
            setattr(self, key, val)
    
    def set_from_read(self, r):
        self.mapped = not r.is_unmapped
        self.quality = r.mapping_quality
        if r.has_tag('AS'):
            self.score = r.get_tag('AS')
        
        if self.mapped:
            self.chrm = r.reference_name
            self.start = r.reference_start
            self.end = r.reference_end
        
            def count_soft_clipped(seq):
                clipped = 0
                for pos in seq:
                    if pos:
                        break
                    else:
                        clipped += 1
                return clipped

            ref_pos = r.get_reference_positions(full_length=True)
            self.clipped_front = count_soft_clipped(ref_pos)
            self.clipped_back = count_soft_clipped(reversed(ref_pos))
    
    def set_trimming(self, u, t, use_edit_distance=True):
        untrimmed = u.query_sequence.upper()
        untrimmed_len = len(untrimmed)
        trimmed = t.query_sequence.upper()
        trimmed_len = len(trimmed)
        
        trimmed_front = 0 if use_edit_distance else -1
        if use_edit_distance and (untrimmed_len > trimmed_len):
            for i in range(untrimmed_len - trimmed_len + 1):
                if untrimmed[i:(i+trimmed_len)] == trimmed:
                    trimmed_front = i
                    break
            else:
                # Since Skewer performs automatic error correction, the trimmed and
                # untrimmed reads may not match, so in that case we find the closest
                # match by Levenshtein distance.
                dist = None
                for i in range(untrimmed_len - trimmed_len + 1):
                    d = editdistance.eval(untrimmed[i:(i+trimmed_len)], trimmed)
                    if not dist:
                        dist = d
                    elif d < dist:
                        trimmed_front = i
                        dist = d
        
        self.trimmed_front = trimmed_front
        self.trimmed_back = untrimmed_len - (trimmed_len + trimmed_front)
    
class TableRow(object):
    def __init__(self, prog, read_name, read_idx, regions=None):
        for key, val in pair_fields:
            setattr(self, key, val)
        self.prog = prog
        self.prog_fields = list(parse_profile(prog))
        self.read_name = read_name
        self.read_idx = read_idx
        self.read1 = TableRead()
        self.read2 = TableRead()
        self.regions = regions
    
    def skip(self, split=False, failed=False, discarded=False):
        self.skipped = True
        if split:
            self.split = True
        if failed:
            self.failed = True
        if discarded:
            self.discarded = True
        if self.regions and not discarded:
            self.regions.skip(self)
    
    def set_from_pair(self, r1, r2):
        assert r1.is_proper_pair == r2.is_proper_pair
        self.proper = r1.is_proper_pair
        # A valid pair is both proper (i.e. both reads are mapped)
        # and read1 and read2 are in opposite orientation
        self.valid = self.proper and (r1.is_reverse != r2.is_reverse)
        self.read1.set_from_read(r1)
        self.read2.set_from_read(r2)
        if self.regions:
            self.regions.set_overlaps(self)
    
    def set_trimming(self, u1, r1, u2, r2, use_edit_distance):
        self.read1.set_trimming(u1, r1, use_edit_distance)
        self.read2.set_trimming(u2, r2, use_edit_distance)
    
    def write(self, w):
        w.writerow(self.prog_fields + [
            getattr(self, key) for key, val in pair_fields[1:]
        ] + [
            getattr(self.read1, key) for key, val in read_fields
        ] + [
            getattr(self.read2, key) for key, val in read_fields
        ])

class Bed(object):
    def __init__(self, path, slop=200):
        from intervaltree import IntervalTree
        from csv import reader
        self.trees = {}
        self.slop = slop
        with open(path, "rt") as bed:
            chrm = None
            ivls = []
            for row in reader(bed, delimiter="\t"):
                if row[0] != chrm:
                    if len(ivls) > 0:
                        self.trees[chrm] = IntervalTree.from_tuples(ivls)
                    chrm = row[0]
                    ivls = []
                ivls.append((
                    max(1, int(row[1]) - slop + 1),
                    int(row[2]) + slop + 1
                ))
            if len(ivls) > 0:
                self.trees[chrm] = IntervalTree.from_tuples(ivls)
    
    def set_overlaps(self, read_pair, min_overlap=1.0):
        for read in (read_pair.read1, read_pair.read2):
            if read.mapped:
                read.in_region = self.overlaps(
                    read.chrm, read.start, read.end, min_overlap)

    def overlaps(self, chrm, start, end, min_overlap=1.0):
        if chrm not in self.trees:
            return 0
        
        tree = self.trees[chrm]
        hits = tree.search(start, end+1)
        
        if len(hits) == 0:
            return 0
        
        length = end - start + 1
        
        for h in hits:
            o = min(end, h.end) - max(start, h.begin) + 1
            l = min(length, h.end - h.begin + 1)
            if o / l >= min_overlap:
                return 1
        
        return 0
    
    def skip(self, read_pair):
        pass
    
    def close(self):
        pass

class Annotations(object):
    def __init__(self, bed_files_or_dir, pattern, untrimmed_name, trimmed_names):
        self.pattern = pattern
        self.annot_beds = {}
        if isinstance(bed_files_or_dir, str):
            self.bed_dir = bed_dir
            self.add(untrimmed_name)
            for name in trimmed_names:
                self.add(name)
        else:
            for path in bed_files_or_dir:
                basename = os.path.basename(path)
                if basename.startswith(untrimmed_name):
                    self.add(untrimmed_name, path)
                    continue
                for name in trimmed_names:
                    if basename.startswith(name):
                        self.add(name, path)
                        break
                else:
                    raise ValueError(
                        "No matching BAM file for BED path {}".format(path))
            if untrimmed_name not in self.annot_beds:
                raise ValueError("No untrimmed BED file")
    
    def add(self, name, path=None):
        if path:
            self.annot_beds[name] = AnnotBed(path)
        else:
            self.annot_beds[name] = AnnotBed(os.path.join(
                self.bed_dir, self.pattern.format(name=name)))
    
    def set_overlaps(self, read_pair, min_overlap=1.0):
        annot = self.annot_beds[read_pair.prog]
        annot.set_overlaps(read_pair, min_overlap)
    
    def skip(self, read_pair):
        annot = self.annot_beds[read_pair.prog]
        annot.skip(read_pair)
    
    def close(self):
        for annots in self.annot_beds.values():
            annots.close()

class AnnotBed(object):
    """
    The result of intersecting a name-sorted BAM file of read mappings with an
    annotation database:
    bam2bed --all-reads --do-not-sort < <bam> | cut -f 1-6 | bedmap \
      --delim '\t' --echo --echo-map-id - <annotations> > out.
    """
    def __init__(self, path):
        import csv
        self.fh = open(path, 'rt')
        self.reader = csv.reader(self.fh, delimiter="\t")
    
    def set_overlaps(self, read_pair, min_overlap=1.0):
        a1, a2 = self.next_annotations(read_pair)
        
        if read_pair.proper:
            # gene names should be semicolon-delimited in the 7th column
            g1 = set(a1[6].split(';'))
            g2 = set(a2[6].split(';'))
            
            # reads are considered "in region" if they are both mapped to
            # the same gene
            intersection = g1 & g2
            if len(intersection) > 0:
                read_pair.read1.in_region = read_pair.read2.in_region = 1
                return
        
        read_pair.read1.in_region = read_pair.read2.in_region = 0
    
    def skip(self, read_pair):
        self.next_annotations(read_pair)
    
    def next_annotations(self, read_pair):
        a1 = next(self.reader)
        a2 = next(self.reader)
        if a1[3] != read_pair.read_name or a2[3] != read_pair.read_name:
            raise Exception("Read {} does not match next annotation {} for sample {}".format(
                read_pair.read_name, a1[3], self.fh.name))
        
        return (a1, a2)
    
    def close(self):
        self.fh.close()

class Hist(object):
    def __init__(self, prog, n_bases):
        # position base histograms
        self.prog = prog
        self.prog_fields = list(parse_profile(prog))
        self.n_bases = n_bases
        self.hists = [
            [
                dict((base, [0] * n_bases) for base in ('A','C','G','T','N'))
                for side in range(2)
            ]
            for read in range(2)
        ]
    
    def add_reads(self, r1, r2):
        # add start and end bases to histogram
        for seq, hist in zip((r1.query_sequence, r2.query_sequence), (self.hists)):
            for i, b in enumerate_range(seq, 0, self.n_bases):
                hist[0][b][i] += 1
            for i, b in enumerate_range(reversed(seq), 0, self.n_bases):
                hist[1][b][i] += 1
    
    def write(self, w):
        for read, read_hists in enumerate(self.hists, 1):
            for side in range(2):
                for base in nuc:
                    for pos, count in enumerate(read_hists[side][base], 1):
                        w.writerow(self.prog_fields + [read, side, pos, base, count])

def summarize(untrimmed, trimmed, ow, hw, mode=None, regions=None,
              n_hist_bases=20, max_reads=None, use_edit_distance=True,
              progress=True):
    progs = ('untrimmed',) + tuple(trimmed.keys())
    hists = dict((prog, Hist(prog, n_hist_bases)) for prog in progs)
    itr = enumerate(untrimmed, 1)
    if progress:
        try:
            import tqdm
            itr = tqdm.tqdm(itr)
        except:
            print("tqdm library not available; not showing progress bar")
    
    for i, (name, u1, u2) in itr:
        assert len(u1) > 0 and len(u2) > 0
        
        rows = dict(untrimmed=TableRow('untrimmed', name, i, regions))
        valid = {}
        fail = False
        skip = False
        
        if len(u1) > 1 or len(u2) > 1:
            rows['untrimmed'].skip(split=True)
            skip = True
        else:
            u1 = u1[0]
            u2 = u2[0]
            if u1.is_qcfail or u2.is_qcfail:
                rows['untrimmed'].skip(failed=True)
                skip = fail = True
            else:
                hists['untrimmed'].add_reads(u1, u2)
                rows['untrimmed'].set_from_pair(u1, u2)
        
        for prog, bam in trimmed.items():
            rows[prog] = TableRow(prog, name, i, regions)
            
            if not bam.finished and bam.peek().query_name == name:
                _, t1, t2 = next(bam)
                assert len(t1) > 0 and len(t2) > 0
                if len(t1) > 1 or len(t2) > 1:
                    rows[prog].skip(split=True, failed=fail)
                elif skip:
                    rows[prog].skip(failed=fail)
                else:
                    t1 = t1[0]
                    t2 = t2[0]
                    hists[prog].add_reads(t1, t2)
                    valid[prog] = (t1, t2)
            else:
                rows[prog].skip(discarded=True, failed=fail)
        
        if len(valid) > 0:
            for prog, (r1, r2) in valid.items():
                row = rows[prog]
                row.set_from_pair(r1, r2)
                row.set_trimming(u1, r1, u2, r2, use_edit_distance)
        
        for prog in progs:
            rows[prog].write(ow)
        
        if max_reads and i >= max_reads:
            break
    
    for h in hists.values():
        h.write(hw)

def main():
    parser = argparse.ArgumentParser()
    parser.set_defaults(command=None)
    parser.add_argument("-i", "--bam-files", nargs="+", default=None)
    parser.add_argument("-d", "--bam-dir")
    parser.add_argument("-x", "--bam-extension", default=".name_sorted.bam")
    parser.add_argument("-p", "--bam-pattern", default=None)
    parser.add_argument("-u", "--untrimmed-name", default="untrimmed")
    parser.add_argument("-o", "--output", default="-")
    parser.add_argument("-H", "--hist", default="trimmed_hists.txt")
    parser.add_argument("-m", "--max-reads", type=int, default=None)
    parser.add_argument("--no-edit-distance", action="store_false",
                        default=True, dest="edit_distance",
                        help="Don't try to match by editdistance.")
    parser.add_argument("--no-progress", action="store_false", dest="progress",
                        default=True)
    sp = parser.add_subparsers()
    
    amplicon = sp.add_parser('amplicon')
    amplicon.set_defaults(command='amplicon')
    amplicon.add_argument(
        "-b", "--bed", default=None,
        help="Sorted .bed file of regions where reads should map.")
    amplicon.add_argument(
        "--min-overlap", type=float, default=1.0,
        help="When a .bed file is specified, this is the minimum "
            "fraction of a mapped read that must overlap a selected "
            "interval for that mapping to be considered valid. (1.0)")
    amplicon.add_argument(
        "--slop", type=int, default=None,
        help="When a .bed file is specified, this is the number of bp each "
            "region is extended. This is often necessary with amplicon data "
            "because enrichment can capture sequences that only paritally "
            "overlap the probes.")
    
    mrna = sp.add_parser('mrna')
    mrna.set_defaults(command='mrna')
    mrna.add_argument(
        "-b", "--bed-files", nargs="+", default=None,
        help="BED file names corresponding to BAM files.")
    mrna.add_argument(
        "-D", "--bed-dir", default=None,
        help="Directory where to find annotation bed files. Defaults to "
            "--bam-dir.")
    mrna.add_argument(
        "-B", "--bed-pattern", default='{name}.bed',
        help="String template for bed file names. \{name\} is replaced with the "
            "BAM file name with extension (--bam-extension) removed.")
    
    args = parser.parse_args()
    
    if args.edit_distance:
        import editdistance
    
    trimmed = {}
    untrimmed = None
    if args.bam_files:
        bam_files = args.bam_files
    else:
        pattern = (args.bam_pattern or "*{}").format(args.bam_extension)
        bam_files = list(glob(os.path.join(args.bam_dir, pattern)))
    for path in bam_files:
        name = os.path.basename(path)[:-len(args.bam_extension)]
        if name == args.untrimmed_name:
            untrimmed = BAMReader(path)
        else:
            trimmed[name] = BAMReader(path)
    if untrimmed is None:
        untrimmed = BAMReader(os.path.join(
            args.bam_dir,
            "{}{}".format(args.untrimmed_name, args.bam_extension)))
    
    regions = None
    if args.command == 'amplicon':
        regions = Bed(args.bed, args.slop or 200)
    elif args.command == 'mrna':
        regions = Annotations(
            args.bed_files or args.bed_dir or args.bam_dir,
            args.bed_pattern,
            args.untrimmed_name,
            trimmed.keys())
    
    try:
        with fileopen(args.output, 'wt') as o, fileopen(args.hist, 'wt') as h:
            ow = csv.writer(o, delimiter="\t")
            ow.writerow(HEADER)
            hw = csv.writer(h, delimiter="\t")
            hw.writerow(prog_header + ['read', 'side', 'pos', 'base', 'count'])
            summarize(untrimmed, trimmed, ow, hw,
                mode=args.command, regions=regions, max_reads=args.max_reads,
                use_edit_distance=args.edit_distance, progress=args.progress)
    finally:
        if untrimmed:
            untrimmed.close()
        for t in trimmed.values():
            t.close()
        if regions:
            regions.close()

if __name__ == "__main__":
    main()

# This is code to compare the mapped reads against the reference.
# from pysam import FastaFile
#parser.add_argument("-r", "--reference",
#    help="Reference sequence. If reads are bisulfite converted, then this must be a "
#         "bisulfite-converted reference.")
#parser.add_argument("-a", "--adapter1", default=ADAPTER1)
#parser.add_argument("-A", "--adapter2", default=ADAPTER2)
#parser.add_argument("-q", "--mapqs", nargs='*', default=(0,10,20,30,40))
# with FastaFile(args.reference) as ref,
# chrm = t.reference_name
# if bisulfite:
#     chrm = ('f','r')[int(t.is_rever)] + chrm
# chrm_len = t.get_reference_length(chrm)
# ref_start = t.reference_start - offset_front
# ref_end = t.reference_end + offset_back
# if ref_start < 0:
#     untrimmed = untrimmed[abs(ref_start):]
#     offset_front += ref_start
#     ref_start = 0
# if ref_end > chrm_len:
#     overhang = chrm_len - ref_end
#     untrimmed = untrimmed[:overhang]
#     offset_back += overhang
#     ref_end = chrm_len
#
# ref_seq = ref.fetch(chrm, ref_start, ref_end)
# assert len(ref_seq) == len(untrimmed)
#
# if bisulfite:
#     untrimmed = bisulfite_convert(untrimmed, r)
#     trimmed = bisulfite_convert(trimmed, r)
#
# def compare_seqs(s1, s2):
#     n = 0
#     for i, (b1, b2) in enumerate(zip(s1, s2)):
#         if b1 == 'N' or b2 == 'N':
#             n += 1
#         elif b1 == b2:
#             n = 0
#         else:
#             break
#     return i - n
#
# if undertrimmed_front == 0 and offset_start > 0:
#     overtrimmed_front = compare_seqs(untrimmed[offset_front:0:-1], ref_seq[offset_front:0:-1])
#
# if undertrimmed_back == 0 and offset_back > 0:
#     overtrimmed_back = compare_seqs(untrimmed[offset_back:], ref_seq[offset_back:]))
#
# def bisulfite_convert(seq, read):
#     if read == 0:
#         return seq.replace('C', 'T')
#     else:
#         return seq.replace('G', 'A')
#
# ADAPTER1="GATCGGAAGAGCACACGTCTGAACTCCAGTCACCAGATCATCTCGTATGCCGTCTTCTGCTTG" # TruSeq index 7
# ADAPTER2="AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT" # TruSeq universal



#!/usr/bin/env python
"""Parse the output of GNU time (gtime), along with a profile name ($item)
and write a tab-delimited row to stdout with the following fields:

1. Program name
2. Number of threads
3. The dataset - either an error rate (for simulated data) or a data type (WGBS or RNA-Seq)
4. Quality cutoff
5. Duration of program execution
6. Max CPU usage
7. Max memory usage
"""
import argparse
from common import fileopen, parse_profile
import re

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", default="-")
    parser.add_argument("-o", "--output", default="-")
    parser.add_argument("-p", "--profile")
    args = parser.parse_args()

    prog, prog2, threads, dataset, qcut = parse_profile(args.profile)
    
    with fileopen(args.input, 'rt') as i:
        lines = [line.strip() for line in i.readlines()]

    cpu = lines[3]
    cpu_match = re.match("Percent of CPU this job got: (\\d+)%", cpu)
    assert cpu_match is not None
    cpu_pct = float(cpu_match.group(1))

    wc_time = lines[4]
    wc_match_prefix = "Elapsed \\(wall clock\\) time \\(h:mm:ss or m:ss\\): "
    wc_match = re.match(wc_match_prefix + "(?:(\\d+)h )?(\\d+)m ([\\d\\.]+)s", wc_time)
    if wc_match is None:
        wc_match = re.match(wc_match_prefix + "(?:(\\d+):)?(\\d+):([\\d\\.]+)", wc_time)
    assert wc_match is not None
    hrs = int(wc_match.group(1) or 0)
    mins = int(wc_match.group(2))
    secs = float(wc_match.group(3))
    duration = ':'.join((
        '{:02d}'.format(hrs),
        '{:02d}'.format(mins),
        '{:0.2f}'.format(secs)))
    duration_secs = (hrs * 3600) + (mins * 60) + secs

    memory = lines[9]
    memory_match = re.match("Maximum resident set size \\(kbytes\\): (\\d+)", memory)
    assert memory_match is not None
    memory_mbytes = int(memory_match.group(1)) / 1000

    with fileopen(args.output, "wt") as out:
        print(prog, prog2, threads, dataset, qcut, duration_secs,
        cpu_pct, memory_mbytes, sep="\t", file=out)

if __name__ == "__main__":
    main()


"""Command-line interface for the detect command.
"""
from atropos.io import STDOUT, STDERR
from atropos.commands.cli import (
    BaseCommandParser, positive, readable_url, writeable_file,
    readwriteable_file, probability)

class CommandParser(BaseCommandParser):
    name = 'detect'
    usage = """
atropos detect -se input.fastq
atropos detect -pe1 in1.fq -pe2 in2.fq
"""
    description = """
Detect adapter sequences directly from read sequences. Use this command if you
are unsure if your data has adapters, or if you know that it has adapters but
you don't know what are the adapter sequences.
"""
    
    def add_command_options(self):
        parser = self.parser
        parser.set_defaults(
            max_reads=10000,
            counter_magnitude="K")
        
        group = self.add_group("Adapter Detection")
        group.add_argument(
            "-d",
            "--detector",
            choices=('known', 'heuristic', 'khmer'), default=None,
            help="Which detector to use. (automatically choose based on other "
                 "options)")
        group.add_argument(
            "-k",
            "--kmer-size",
            type=positive(), default=12,
            help="Size of k-mer used to scan reads for adapter sequences. (12)")
        group.add_argument(
            "-e",
            "--past-end-bases", nargs="*", default=('A',),
            help="On Illumina, long runs of A (and sometimes other bases) "
                 "can signify that the sequencer has read past the end of a "
                 "fragment that is shorter than the read length + adapter "
                 "length. Those bases will be removed from any sequencers "
                 "before looking for matching contaminants. Can also be a "
                 "regular expression.")
        group.add_argument(
            "-i",
            "--include-contaminants",
            choices=('all','known','unknown'), default='all',
            help="What conaminants to search for: all, only known "
                 "adapters/contaminants ('known'), or only unknown "
                 "contaminants ('unknown'). (all)")
        group.add_argument(
            "-x",
            "--known-contaminant",
            action="append", dest='known_adapter', default=None,
            help="Pass known contaminants in on the commandline as "
                 "'name=sequence'. Can be specified multiple times.")
        group.add_argument(
            "-F",
            "--known-contaminants-file",
            type=readable_url, action="append", dest='known_adapters_file',
            default=None,
            help="Points to FASTA File or URL with known contaminants.")
        group.add_argument(
            "--no-default-contaminants",
            action="store_false", dest="default_adapters", default=True,
            help="Don't fetch the default contaminant list (which is currently "
                 "stored as a GitHub gist).")
        group.add_argument(
            "--contaminant-cache-file",
            type=readwriteable_file, dest='adapter_cache_file',
            default='.adapters',
            help="File where known contaminant sequences will be cached, "
                 "unless --no-cache-contaminants is set.")
        group.add_argument(
            "--no-cache-contaminants",
            action="store_false", dest="cache_adapters", default=True,
            help="Don't cache contaminant list as '.contaminants' in the "
                 "working directory.")
        
        group = self.add_group("Known Detector Options")
        group.add_argument(
            "--min-kmer-match-frac",
            type=probability, default=0.5,
            help="Minimum fraction of contaminant kmers that must be found "
                 "in a read sequence to be considered a match.")
        
        group = self.add_group("Heuristic Detector Options")
        group.add_argument(
            "--min-frequency",
            type=probability, default=0.001,
            help="Minimum frequency required to retain a k-mer.")
        group.add_argument(
            "--min-contaminant-match-frac",
            type=probability, default=0.9,
            help="Minimum fraction of nucleotides that must align for a "
                 "detected contaminant to match a known adapter sequence.")
        
        group = self.add_group("Output")
        group.add_argument(
            "-o",
            "--output",
            type=writeable_file, default=STDOUT, metavar="FILE",
            help="File in which to write the summary of detected adapters. "
                 "(stdout)")
        group.add_argument(
            "-O",
            "--output-formats",
            nargs="*", choices=("txt", "fasta", "json", "yaml", "pickle"), 
            default=None, metavar="FORMAT", dest="report_formats",
            help="Report type(s) to generate. If multiple, '--output' "
                 "is treated as a prefix and the appropriate extensions are "
                 "appended. If unspecified, the format is guessed from the "
                 "file extension. Supported formats are: txt, json, yaml, "
                 "pickle, fasta. Additional arguments for fasta output are "
                 "provided via the '--fasta' option. See the documentation "
                 "for a full description of the structured output "
                 "(json/yaml/pickle formats).")
        group.add_argument(
            "--fasta",
            nargs="*", choices=("union", "perinput"), default=None, 
            metavar="OPTIONS",
            help="Additional arguments for fasta output. Adds 'fasta' to the "
                 "list of output formats if not already specified. Options "
                 "are: perinput=generate one output file per input file, "
                 "union=generate a single output file with all sequences "
                 "merged.")
        group.add_argument(
            "-m",
            "--max-adapters",
            type=positive(), default=None,
            help="The maximum number of candidate adapters to report. "
                 "(report all)")
    
    def validate_command_options(self, options):
        options.report_file = options.output
        is_std = options.report_file in (STDOUT, STDERR)
        if options.fasta:
            if is_std and 'perinput' in options.fasta:
                self.parser.error("Per-input fasta cannot be written to stdout")
            if not options.report_formats:
                options.report_formats = ['fasta']
            elif 'fasta' not in options.report_formats:
                options.report_formats = list(options.report_formats) + ['fasta']
        elif is_std and options.report_formats and 'fasta' in options.report_formats:
            options.fasta = ['union']



#!/usr/bin/env python
"""Summarizes trimming accuracy for simulated reads."""

from argparse import ArgumentParser
import csv
import os
from common import *

# Either of these will work
from editdistance import eval as editdistance
#from Levenshtein import distance as editdistance

summary_fields = (
    "program", "program2", "threads", "dataset", "qcut", "retained reads", 
    "mismatch reads", "discarded reads", "total reads", "reads with adapters", 
    "retained reads with adapters", "non-adapter reads trimmed", 
    "adapter reads untrimmed", "adapter reads undertrimmed", 
    "adapter reads overtrimmed", "total ref bases", "total ref edit distance", 
    "total adapter bases", "total retained adapter bases", 
    "total adapter edit dist", "overtrimmed bases", "undertrimmed bases"
)

def summarize_accuracy(aln_iter, read_iter, w, read_length, adapters):
    adapter_lengths = [len(adapters[i]) for i in (0,1)]
    
    debug = False
    
    def summarize_alignment(a):
        ref_seq = a[5]
        if debug: print(ref_seq)
        ref_len = len(ref_seq)
        if debug: print(ref_len)
        read_seq = a[6]
        if debug: print(read_seq)
        read_len = len(read_seq)
        if debug: print(read_len)
        
        read_ref = read_seq[:ref_len]
        
        ref_ins = [i for i in range(len(ref_seq)) if ref_seq[i] == '-']
        expected_read = "".join(b for b in read_ref if b != '-')
        expected_read_len = len(expected_read)
        ref_del = ref_len - expected_read_len
        
        has_adapter = ref_len < read_len
        adapter_seq = []
        adapter_len = adapter_ins = adapter_del = polyA = 0
        if debug: print(has_adapter)
        
        if has_adapter:
            for b in read_seq[ref_len:]:
                if adapter_len >= adapter_lengths[i] and b == 'A':
                    polyA += 1
                else:
                    if b == '-':
                        adapter_del += 1
                    else:
                        adapter_seq.append(b)
                        adapter_len += 1
            
            adapter_ref_len = adapter_len + adapter_del
            if adapter_ref_len > adapter_lengths[i]:
                adapter_ins = adapter_ref_len - adapter_lengths[i]
        
        ref_edit_dist = editdistance(ref_seq, read_ref)
        adapter_edit_dist = editdistance(adapters[i][:adapter_len], "".join(adapter_seq))
        return [expected_read, expected_read_len,
            (ref_len, ref_edit_dist),
            (int(has_adapter), adapter_len, adapter_edit_dist, adapter_ins, adapter_del, polyA)
        ]
    
    cache = {}
    
    total_ref_bp = 0
    total_ref_edit_dist = 0
    total_adapter_bp = 0
    total_adapter_edit_dist = 0
    overtrimmed_bp = 0
    undertrimmed_bp = 0
    raw_trimmed_mismatch = 0
    num_adapter_reads = 0
    adapter_reads_untrimmed = 0
    adapter_reads_undertrimmed = 0
    adapter_reads_overtrimmed = 0
    non_adapter_reads_trimmed = 0
    num_aln = 0
    num_discarded_adapter_reads = 0
    num_discarded_adapter_bp = 0
    
    for num_reads, reads in enumerate(read_iter, 1):
        read_id = reads[0][0]
        aln = None
        
        assert read_id == reads[1][0], "Read IDs differ - {} != {}".format(read_id, reads[1][0])
        assert int(reads[0][1]) == 1 and int(reads[1][1]) == 2, "Mate identifiers are incorrect for {}".format(read_id)
        if read_id not in cache:
            for aln in aln_iter:
                num_aln += 1
                if read_id == aln[0][0]:
                    break
                else:
                    cache[aln[0][0]] = aln
        else:
            aln = cache.pop(read_id)
        
        if debug: print(reads)
        if debug: print(aln)
        
        if aln is None:
            raise Exception("No alignment for read {}".format(read_id))
        
        for i in (0, 1):
            expected_read, expected_read_len, ref_info, adapter_info = summarize_alignment(aln[i])
            total_ref_bp += ref_info[0]
            total_ref_edit_dist += ref_info[1]
            has_adapter = adapter_info[0] == 1
            total_adapter_bp += adapter_info[1]
            total_adapter_edit_dist += adapter_info[2]
            
            if has_adapter:
                num_adapter_reads += 1
            
            r = reads[i]
            trimmed_len = len(r[2])
            
            if debug: print(trimmed_len)
            if debug: print(r[2])
            if debug: print(expected_read)
            
            status = 'OK'
            common_len = min(trimmed_len, expected_read_len)
            if expected_read_len > trimmed_len:
                overtrimmed_bp += expected_read_len - trimmed_len
                status = 'OVERTRIMMED'
                if has_adapter:
                    adapter_reads_overtrimmed += 1
                else:
                    non_adapter_reads_trimmed += 1
            elif expected_read_len < trimmed_len:
                undertrimmed_bp += trimmed_len - expected_read_len
                status = 'UNDERTRIMMED'
                if not has_adapter:
                    raise Exception("Expected read length is less than trimmed "
                                    "length for read without an adapter: {}".format(read_id))
                elif trimmed_len == read_length:
                    adapter_reads_untrimmed += 1
                else:
                    adapter_reads_undertrimmed += 1
            if r[2][:common_len] != expected_read[:common_len]:
                # This happens with Skewer results due to automatic error correction
                raw_trimmed_mismatch += 1
                status = 'MISMATCH'
            
            w.writerow((read_id, i+1, expected_read_len, trimmed_len, status) + ref_info + adapter_info)

    # all remaining alignments represent reads that were discarded
    
    def handle_discarded(aln):
        read_id = aln[0][0]
        for i in (0, 1):
            expected_read, expected_read_len, ref_info, adapter_info = summarize_alignment(aln[i])
            w.writerow((read_id, i+1, expected_read_len, '', 'DISCARDED') + ref_info + adapter_info)
            if adapter_info[0] == 1:
                nonlocal num_discarded_adapter_reads, num_discarded_adapter_bp
                num_discarded_adapter_reads += 1
                num_discarded_adapter_bp += adapter_info[1]
    
    num_discarded = len(cache)
    for aln in cache.values():
        handle_discarded(aln)
    for aln in aln_iter:
        handle_discarded(aln)
        num_discarded += 1
    
    return (
        num_reads,
        raw_trimmed_mismatch,
        num_discarded,
        num_reads + num_discarded,
        num_adapter_reads + num_discarded_adapter_reads,
        num_adapter_reads,
        non_adapter_reads_trimmed,
        adapter_reads_untrimmed,
        adapter_reads_undertrimmed,
        adapter_reads_overtrimmed,
        total_ref_bp,
        total_ref_edit_dist,
        total_adapter_bp + num_discarded_adapter_bp,
        total_adapter_bp,
        total_adapter_edit_dist,
        overtrimmed_bp,
        undertrimmed_bp
    )

def main():
    parser = ArgumentParser()
    parser.add_argument('-a1', '--aln1', help=".aln file associated with simulated read1")
    parser.add_argument('-a2', '--aln2', help=".aln file associated with simulated read2")
    parser.add_argument('-r1', '--reads1', help="trimmed fastq file read1")
    parser.add_argument('-r2', '--reads2', help="trimmed fastq file read2")
    parser.add_argument('-l', '--read-length', type=int, default=125)
    parser.add_argument('-o', '--output', default='-')
    parser.add_argument('-s', '--summary', default='-')
    parser.add_argument('-t', '--table', default=None)
    parser.add_argument("-p", "--profile", default=None)
    parser.add_argument("--adapters", nargs=2, default=DEFAULT_ADAPTERS)
    parser.add_argument("--no-progress", action="store_true", default=False)
    args = parser.parse_args()
    
    profile = parse_profile(args.profile)
    
    with fileopen(args.aln1, 'rt') as a1, fileopen(args.aln2, 'rt') as a2:
        aln_pair_iterator = zip(aln_iterator(a1), aln_iterator(a2))
        
        with fileopen(args.reads1, 'rt') as r1, fileopen(args.reads2, 'rt') as r2:
            read_pair_iterator = zip(fq_iterator(r1, 1, True), fq_iterator(r2, 2, True))
            
            if not args.no_progress:
                try:
                    import tqdm
                    aln_pair_iterator = iter(tqdm.tqdm(aln_pair_iterator))
                except:
                    print("tqdm library is required for a progress bar")
            
            with fileopen(args.output, 'wt') as o:
                w = csv.writer(o, delimiter="\t")
                w.writerow((
                    'read_id','mate','expected_len','actual_len','status','has_adapter',
                    'adapter_len','adapter_edit_dist','adapter_ins','adapter_del','polyA'))
                summary = summarize_accuracy(
                    aln_pair_iterator, read_pair_iterator, w, args.read_length, 
                    args.adapters)
            
            with fileopen(args.summary, 'wt') as s:
                for field, value in zip(("{} " + field for field in summary_fields), summary):
                    print(field.format(value), file=s)
            
            if args.table:
                with fileopen(args.table, "wt") as t:
                    w = csv.writer(t, delimiter="\t")
                    w.writerow(profile + summary)

if __name__ == "__main__":
    main()



# coding: utf-8
from pytest import raises
from collections import defaultdict
import random
import sys
import os
from io import StringIO
import shutil
from textwrap import dedent
from tempfile import mkdtemp
from atropos.io import xopen, open_output
from atropos.io.seqio import (
    Sequence,
    ColorspaceSequence,
    FormatError,
    FastaReader,
    FastqReader,
    FastaQualReader,
    InterleavedSequenceReader,
    FastaFormat,
    FastqFormat,
    InterleavedFormatter,
    get_format,
    open_reader as openseq,
    sequence_names_match,
)
from atropos.util import ALPHABETS
from .utils import temporary_path

# files tests/data/simple.fast{q,a}
simple_fastq = [
    Sequence("first_sequence", "SEQUENCE1", ":6;;8<=:<"),
    Sequence("second_sequence", "SEQUENCE2", "83<??:(61"),
]
simple_fasta = [Sequence(x.name, x.sequence, None) for x in simple_fastq]


class TestAlphabet:

    def test_alphabet(self):
        alphabet = ALPHABETS['dna']
        for base in ('A', 'C', 'G', 'T', 'N'):
            assert base in alphabet
        assert 'X' not in alphabet
        assert alphabet.resolve('X') == 'N'


class TestSequence:

    def test_too_many_qualities(self):
        with raises(FormatError):
            Sequence(name="name", sequence="ACGT", qualities="#####")

    def test_too_many_qualities_colorspace(self):
        with raises(FormatError):
            ColorspaceSequence(name="name", sequence="T0123", qualities="#####")

    def test_invalid_primer(self):
        with raises(FormatError):
            ColorspaceSequence(name="name", sequence="K0123", qualities="####")


class TestFastaReader:

    def test(self):
        with FastaReader("tests/data/simple.fasta") as f:
            reads = list(f)
        assert reads == simple_fasta
        fasta = StringIO(">first_sequence\nSEQUENCE1\n>second_sequence\nSEQUENCE2\n")
        reads = list(FastaReader(fasta))
        assert reads == simple_fasta

    def test_with_comments(self):
        fasta = StringIO(
            dedent("""
            # a comment
            # another one
            >first_sequence
            SEQUENCE1
            >second_sequence
            SEQUENCE2
            """)
        )
        reads = list(FastaReader(fasta))
        assert reads == simple_fasta

    def test_wrong_format(self):
        with raises(FormatError):
            fasta = StringIO(
                dedent("""
                # a comment
                # another one
                unexpected
                >first_sequence
                SEQUENCE1
                >second_sequence
                SEQUENCE2
                """)
            )
            reads = list(FastaReader(fasta))

    def test_fastareader_keeplinebreaks(self):
        with FastaReader("tests/data/simple.fasta", keep_linebreaks=True) as f:
            reads = list(f)
        assert reads[0] == simple_fasta[0]
        assert reads[1].sequence == 'SEQUEN\nCE2'

    def test_context_manager(self):
        filename = "tests/data/simple.fasta"
        with open(filename) as f:
            assert not f.closed
            reads = list(openseq(f))
            assert not f.closed
        assert f.closed
        with FastaReader(filename) as sr:
            tmp_sr = sr
            assert not sr._file.closed
            reads = list(sr)
            assert not sr._file.closed
        assert tmp_sr._file is None
        # Open it a second time
        with FastaReader(filename) as sr:
            pass


class TestFastqReader:

    def test_fastqreader(self):
        with FastqReader("tests/data/simple.fastq") as f:
            reads = list(f)
        assert reads == simple_fastq

    def test_fastqreader_dos(self):
        with FastqReader("tests/data/dos.fastq") as f:
            dos_reads = list(f)
        with FastqReader("tests/data/small.fastq") as f:
            unix_reads = list(f)
        assert dos_reads == unix_reads

    def test_fastq_wrongformat(self):
        with raises(FormatError), FastqReader("tests/data/withplus.fastq") as f:
            reads = list(f)

    def test_fastq_incomplete(self):
        fastq = StringIO("@name\nACGT+\n")
        with raises(FormatError), FastqReader(fastq) as fq:
            list(fq)

    def test_context_manager(self):
        filename = "tests/data/simple.fastq"
        with open(filename) as f:
            assert not f.closed
            reads = list(openseq(f))
            assert not f.closed
        assert f.closed
        with FastqReader(filename) as sr:
            tmp_sr = sr
            assert not sr._file.closed
            reads = list(sr)
            assert not sr._file.closed
        assert tmp_sr._file is None

    def test_alphabet(self):
        filename = "tests/data/bad_bases.fq"
        with FastqReader(filename, alphabet=ALPHABETS['dna']) as f:
            reads = list(f)
            assert reads[0].sequence == 'ACGNGGACT'
            assert reads[1].sequence == 'CGGACNNNC'


class TestFastaQualReader:

    def test_mismatching_read_names(self):
        with raises(FormatError):
            fasta = StringIO(">name\nACG")
            qual = StringIO(">nome\n3 5 7")
            list(FastaQualReader(fasta, qual))

    def test_invalid_quality_value(self):
        with raises(FormatError):
            fasta = StringIO(">name\nACG")
            qual = StringIO(">name\n3 xx 7")
            list(FastaQualReader(fasta, qual))


class TestSeqioOpen:

    def setup(self):
        self._tmpdir = mkdtemp()

    def teardown(self):
        shutil.rmtree(self._tmpdir)

    def test_sequence_reader(self):
        # test the autodetection
        with openseq("tests/data/simple.fastq") as f:
            reads = list(f)
        assert reads == simple_fastq
        with openseq("tests/data/simple.fasta") as f:
            reads = list(f)
        assert reads == simple_fasta
        with open("tests/data/simple.fastq") as f:
            reads = list(openseq(f))
        assert reads == simple_fastq
        # make the name attribute unavailable
        f = StringIO(open("tests/data/simple.fastq").read())
        reads = list(openseq(f))
        assert reads == simple_fastq
        f = StringIO(open("tests/data/simple.fasta").read())
        reads = list(openseq(f))
        assert reads == simple_fasta

    def test_autodetect_fasta_format(self):
        path = os.path.join(self._tmpdir, 'tmp.fasta')
        fmt = get_format(path)
        assert isinstance(fmt, FastaFormat)
        with open_output(path, "w") as f:
            for seq in simple_fasta:
                f.write(fmt.format(seq))
        assert list(openseq(path)) == simple_fasta

    def test_write_qualities_to_fasta(self):
        path = os.path.join(self._tmpdir, 'tmp.fasta')
        fmt = get_format(path, qualities=True)
        assert isinstance(fmt, FastaFormat)
        with open_output(path, "w") as f:
            for seq in simple_fastq:
                f.write(fmt.format(seq))
        assert list(openseq(path)) == simple_fasta

    def test_autodetect_fastq_format(self):
        path = os.path.join(self._tmpdir, 'tmp.fastq')
        fmt = get_format(path)
        with open_output(path, "w") as f:
            for seq in simple_fastq:
                f.write(fmt.format(seq))
        assert list(openseq(path)) == simple_fastq

    def test_fastq_qualities_missing(self):
        with raises(ValueError):
            path = os.path.join(self._tmpdir, 'tmp.fastq')
            get_format(path, qualities=False)


class TestInterleavedReader:

    def test(self):
        expected = [
            (
                Sequence('read1/1 some text', 'TTATTTGTCTCCAGC', '##HHHHHHHHHHHHH'),
                Sequence('read1/2 other text', 'GCTGGAGACAAATAA', 'HHHHHHHHHHHHHHH'),
            ),
            (
                Sequence('read3/1', 'CCAACTTGATATTAATAACA', 'HHHHHHHHHHHHHHHHHHHH'),
                Sequence('read3/2', 'TGTTATTAATATCAAGTTGG', '#HHHHHHHHHHHHHHHHHHH'),
            ),
        ]
        reads = list(InterleavedSequenceReader("tests/cut/interleaved.fastq"))
        for (r1, r2), (e1, e2) in zip(reads, expected):
            print(r1, r2, e1, e2)
        assert reads == expected
        with openseq("tests/cut/interleaved.fastq", interleaved=True) as f:
            reads = list(f)
        assert reads == expected

    def test_missing_partner(self):
        with raises(FormatError):
            s = StringIO('@r1\nACG\n+\nHHH')
            list(InterleavedSequenceReader(s))

    def test_incorrectly_paired(self):
        with raises(FormatError):
            s = StringIO('@r1/1\nACG\n+\nHHH\n@wrong_name\nTTT\n+\nHHH')
            list(InterleavedSequenceReader(s))


class TestFastaWriter:

    def setup(self):
        self._tmpdir = mkdtemp()
        self.path = os.path.join(self._tmpdir, 'tmp.fasta')

    def teardown(self):
        shutil.rmtree(self._tmpdir)

    def test(self):
        fmt = FastaFormat()
        with open_output(self.path, "w") as fw:
            fw.write(fmt.format_entry("name", "CCATA"))
            fw.write(fmt.format_entry("name2", "HELLO"))
        with open(self.path) as t:
            assert t.read() == '>name\nCCATA\n>name2\nHELLO\n'

    def test_linelength(self):
        fmt = FastaFormat(line_length=3)
        with open_output(self.path, "w") as fw:
            fw.write(fmt.format_entry("r1", "ACG"))
            fw.write(fmt.format_entry("r2", "CCAT"))
            fw.write(fmt.format_entry("r3", "TACCAG"))
        with open(self.path) as t:
            x = t.read()
            print(x)
            assert x == '>r1\nACG\n>r2\nCCA\nT\n>r3\nTAC\nCAG\n'

    def test_write_sequence_object(self):
        fmt = FastaFormat()
        with open_output(self.path, "w") as fw:
            fw.write(fmt.format(Sequence("name", "CCATA")))
            fw.write(fmt.format(Sequence("name2", "HELLO")))
        with open(self.path) as t:
            assert t.read() == '>name\nCCATA\n>name2\nHELLO\n'

    def test_write_zero_length_sequence(self):
        fmt = FastaFormat()
        s = fmt.format_entry("name", "")
        assert s == '>name\n\n', '{0!r}'.format(s)


class TestFastqWriter:

    def setup(self):
        self._tmpdir = mkdtemp()
        self.path = os.path.join(self._tmpdir, 'tmp.fastq')

    def teardown(self):
        shutil.rmtree(self._tmpdir)

    def test(self):
        fmt = FastqFormat()
        with open_output(self.path, "w") as fw:
            fw.write(fmt.format_entry("name", "CCATA", "!#!#!"))
            fw.write(fmt.format_entry("name2", "HELLO", "&&&!&&"))
        with open(self.path) as t:
            assert t.read() == '@name\nCCATA\n+\n!#!#!\n@name2\nHELLO\n+\n&&&!&&\n'

    def test_twoheaders(self):
        fmt = FastqFormat()
        with open_output(self.path, "w") as fw:
            fw.write(fmt.format(Sequence("name", "CCATA", "!#!#!", name2="name")))
            fw.write(fmt.format(Sequence("name2", "HELLO", "&&&!&", name2="name2")))
        with open(self.path) as t:
            assert t.read(
            ) == '@name\nCCATA\n+name\n!#!#!\n@name2\nHELLO\n+name2\n&&&!&\n'


class TestInterleavedWriter:

    def test(self):
        reads = [
            (
                Sequence('A/1 comment', 'TTA', '##H'),
                Sequence('A/2 comment', 'GCT', 'HH#'),
            ),
            (Sequence('B/1', 'CC', 'HH'), Sequence('B/2', 'TG', '#H')),
        ]
        fmt = InterleavedFormatter(FastqFormat(), "foo")
        result = defaultdict(lambda: [])
        for read1, read2 in reads:
            fmt.format(result, read1, read2)
        assert fmt.written == 2
        assert fmt.read1_bp == 5
        assert fmt.read2_bp == 5
        assert "foo" in result
        assert "".join(
            result["foo"]
        ) == '@A/1 comment\nTTA\n+\n##H\n@A/2 comment\nGCT\n+\nHH#\n@B/1\nCC\n+\nHH\n@B/2\nTG\n+\n#H\n'


class TestPairedSequenceReader:

    def test_sequence_names_match(self):

        def match(name1, name2):
            seq1 = Sequence(name1, 'ACGT')
            seq2 = Sequence(name2, 'AACC')
            return sequence_names_match(seq1, seq2)

        assert match('abc', 'abc')
        assert match('abc/1', 'abc/2')
        assert match('abc.1', 'abc.2')
        assert match('abc1', 'abc2')
        assert not match('abc', 'xyz')


def create_truncated_file(path):
    # Random text
    text = ''.join(random.choice('ABCDEFGHIJKLMNOPQRSTUVWXYZ') for _ in range(200))
    f = xopen(path, 'w')
    f.write(text)
    f.close()
    f = open(path, 'a')
    f.truncate(os.stat(path).st_size - 10)
    f.close()


# Disable these tests in Python 3.2 and 3.3
if not ((3, 2) <= sys.version_info[:2] <= (3, 3)):

    def test_truncated_gz():
        with raises(EOFError), temporary_path('truncated.gz') as path:
            create_truncated_file(path)
            f = xopen(path, 'r')
            f.read()
            f.close()

    def test_truncated_gz_iter():
        with raises(EOFError), temporary_path('truncated.gz') as path:
            create_truncated_file(path)
            f = xopen(path, 'r', use_system=False)  # work around bug in py3.4
            for line in f:
                pass
            f.close()



"""Estimate the empircal error rate.
"""
from collections import defaultdict
import csv
import re
from atropos import AtroposError
from atropos.commands.base import (
    BaseCommandRunner, Pipeline, SingleEndPipelineMixin, PairedEndPipelineMixin)
from atropos.io import open_output
from atropos.util import qual2prob, run_interruptible

class CommandRunner(BaseCommandRunner):
    name = 'error'
    
    def __call__(self):
        if not self.delivers_qualities:
            raise ValueError(
                "Cannot estimate error rate without base qualities")
        
        if self.algorithm == 'quality':
            estimator_class = BaseQualityErrorEstimator
        elif self.algorithm == 'shadow':
            estimator_class = ShadowRegressionErrorEstimator
        
        estimator_args = dict(max_read_len=self.max_bases)
        if self.paired:
            estimator = PairedErrorEstimator(
                estimator_class=estimator_class, **estimator_args)
        else:
            estimator = estimator_class(**estimator_args)
        
        self.summary['errorrate'] = estimator_args
        
        # currently only single-threaded operation is supproted
        self.summary.update(mode='serial', threads=1)
        return run_interruptible(estimator, self, raise_on_error=True)

class ErrorEstimator(SingleEndPipelineMixin, Pipeline):
    """Base class for error estimators.
    """
    def __init__(self, max_read_len):
        super().__init__()
        self.total_len = 0
        self.max_read_len = max_read_len
    
    def handle_reads(self, context, read1, read2=None):
        raise NotImplementedError()
    
    def estimate(self):
        """Returns an estimate of the error rate.
        """
        raise NotImplementedError()
    
    def finish(self, summary, **kwargs):
        super().finish(summary)
        estimate, details = self.estimate()
        summary['errorrate'].update(
            estimate=(estimate,), 
            total_len=(self.total_len,),
            details=(details,))

class BaseQualityErrorEstimator(ErrorEstimator):
    """Simple error estimation using base qualities. It is well-known that base
    qualities significantly overestimate true error rates, so take the estimate
    with a grain of salt.
    """
    def __init__(self, max_read_len=None):
        super().__init__(max_read_len)
        self.total_qual = 0.0
    
    def handle_reads(self, context, read1, read2=None):
        quals = read1.qualities
        readlen = len(quals)
        if self.max_read_len and self.max_read_len < readlen:
            readlen = self.max_read_len
            quals = quals[:readlen]
        self.total_qual += sum(qual2prob(qchar) for qchar in quals)
        self.total_len += readlen
    
    def estimate(self):
        return (self.total_qual / self.total_len, None)

# Error estimation using shadow counts
# DEPRECATED
# TODO: Replace with Zhu et al. cubic splines method or something faster
# https://bmcbioinformatics.biomedcentral.com/articles/10.1186/s12859-016-1052-3

FILTER_RE = re.compile("A+|C+|G+|T+|.*N.*")

SHADOW_REGRESSION_R_SCRIPT = """
library(ShadowRegression)
errorRates = getErrorRates("{reads}", type="{method}")
write.table(errorRates$perReadER, "{per_read}", sep="\t", quote=F, col.names=F,
            row.names=T)
write.table(errorRates$cycleER, "{per_cycle}", sep="\t", quote=F, col.names=F,
            row.names=T)
"""

class ShadowRegressionErrorEstimator(ErrorEstimator):
    """Re-implementation of the shadow regression method described in:
    Wang et al., "Estimation of sequencing error rates in short reads",
        BMC Bioinformatics 2012 13:185, DOI: 10.1186/1471-2105-13-185
    
    Args:
        method: The differences that are considered in the error rate
            calculation; sub = substitutions, indel = insertions and
            deletions; all = both substitutions and indels.
        max_read_len: The maximum number of bases (starting from the 5' end)
            to consider from each read.
    """
    def __init__(self, method='sub', max_read_len=None, rscript_exe="Rscript"):
        super().__init__(max_read_len)
        self.seqs = defaultdict(lambda: 0)
        self.method = method
        self.rscript_exe = rscript_exe
    
    def handle_reads(self, context, read1, read2=None):
        seq = read1.sequence
        readlen = len(seq)
        if self.max_read_len and self.max_read_len < readlen:
            readlen = self.max_read_len
            seq = seq[:readlen]
        if FILTER_RE.fullmatch(seq):
            return
        self.seqs[seq] += 1
        self.total_len += readlen
    
    def estimate(self):
        # This is a temporary solution that requires R and the
        # ShadowRegression package. Eventually, this will be
        # replaced with a pure-python implementation.
        import os
        import subprocess
        import tempfile
        
        tempfiles = (tempfile.mkstemp()[1] for i in range(4))
        read_counts, per_read, per_cycle, script_file = tempfiles
        try:
            # write counts to a file
            with open(read_counts, 'wt') as out:
                self._write_read_counts(out)
            
            # execute R script
            script = SHADOW_REGRESSION_R_SCRIPT.format(
                reads=read_counts, method=self.method, per_read=per_read,
                per_cycle=per_cycle)
            with open(script_file, 'wt') as out:
                out.write(script)
            proc = subprocess.Popen(
                [self.rscript_exe, "--vanilla", script_file],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            with proc:
                stdout, stderr = proc.communicate()
                if proc.returncode != 0:
                    raise AtroposError(
                        "R script failed: rc={}; stdout={}; stderr={}".format(
                            proc.returncode, stdout, stderr))
            
            # read the results
            with open(per_read, 'rt') as i:
                reader = csv.reader(i, delimiter="\t")
                per_read_error = dict(reader)
                if len(per_read_error) != 4:
                    raise AtroposError("Invalid output from R script")
            with open(per_cycle, 'rt') as i:
                reader = csv.reader(i, delimiter="\t")
                per_cycle_error = list(row[0:3] for row in reader)
                if not per_cycle_error:
                    raise AtroposError("Invalid output from R script")
            
            return (
                per_read_error["error rate"],
                dict(per_read=per_read_error, per_cycle=per_cycle_error))
        finally:
            for path in tempfiles:
                os.remove(path)
    
    def _write_read_counts(self, fileobj):
        writer = csv.writer(fileobj, delimiter=" ")
        writer.writerows(sorted(
            self.seqs.items(), reverse=True, key=lambda i: i[1]))

class PairedErrorEstimator(PairedEndPipelineMixin, Pipeline):
    """Estimator for a pair of input files.
    """
    def __init__(
            self, estimator_class=BaseQualityErrorEstimator, **kwargs):
        super().__init__()
        self.estimator1 = estimator_class(**kwargs)
        self.estimator2 = estimator_class(**kwargs)
    
    def handle_reads(self, context, read1, read2):
        self.estimator1.handle_reads(context, read1)
        self.estimator2.handle_reads(context, read2)
    
    def finish(self, summary, **kwargs):
        """Estimate error rates.
        
        Returns:
            Tuple (read1_estimate, read2_estimate).
        """
        super().finish(summary)
        estimate1, details1 = self.estimator1.estimate()
        estimate2, details2 = self.estimator2.estimate()
        summary['errorrate'].update(
            estimate=(estimate1, estimate2),
            total_len=(self.estimator1.total_len, self.estimator2.total_len),
            details=(details1, details2))



"""Widely useful utility methods.
"""
from collections import OrderedDict, Iterable, Sequence
from datetime import datetime
import errno
import functools
import logging
import math
from numbers import Number
import time
from atropos import AtroposError

# TODO: the nucleotide table should be implemented as an alphabet.

class NotInAlphabetError(Exception):
    def __init__(self, character):
        super().__init__()
        self.character = character

class Alphabet():
    def __init__(self, valid_characters, default_character):
        if not isinstance(valid_characters, set):
            valid_characters = set(valid_characters)
        if not default_character in valid_characters:
            valid_characters.add(default_character)
        self.valid_characters = valid_characters
        self.default_character = default_character

    def __contains__(self, character):
        return character in self.valid_characters

    def validate(self, character):
        """Raises NotInAlphabetError if the character is not in the alphabet.
        """
        if not character in self:
            raise NotInAlphabetError(character)

    def validate_string(self, string):
        """Raises NotInAlphabetError if any character in 'string' is not in the
        alphabet.
        """
        for character in string:
            self.validate(character)

    def resolve(self, character):
        """Returns 'character' if it's in the alphabet, otherwise the
        alphabet's default character.
        """
        if character in self.valid_characters:
            return character
        else:
            return self.default_character

    def resolve_string(self, string):
        """Returns a new string with any non-alphabet characters replaced
        with the default character.
        """
        return "".join(self.resolve(c) for c in string)

ALPHABETS = dict(
    dna=Alphabet('ACGT', 'N'),
    iso=None,
    colorspace=Alphabet('0123', None)
)

def build_iso_nucleotide_table():
    """Generate a dict mapping ISO nucleotide characters to their complements,
    in both upper and lower case.
    """
    nuc = {
        'A' : 'T',
        'C' : 'G',
        'R' : 'Y',
        'S' : 'S',
        'W' : 'W',
        'K' : 'M',
        'B' : 'V',
        'D' : 'H',
        'N' : 'N'
    }
    for base, comp in tuple(nuc.items()):
        nuc[comp] = base
        nuc[base.lower()] = comp.lower()
        nuc[comp.lower()] = base.lower()
    return nuc

BASE_COMPLEMENTS = build_iso_nucleotide_table()

IUPAC_BASES = frozenset(('X',) + tuple(BASE_COMPLEMENTS.keys()))
"""Valid IUPAC bases, plus 'X'"""

GC_BASES = frozenset('CGRYSKMBDHVN')
"""IUPAC bases that include C or G."""

MAGNITUDE = dict(
    G=1E9,
    M=1E6,
    K=1E3
)

LOG2 = math.log(2)

class RandomMatchProbability(object):
    """Class for computing random match probability for DNA sequences based on
    binomial expectation. Maintains a cache of factorials to speed computation.

    Args:
        init_size: Initial cache size.
    """
    def __init__(self, init_size=150):
        self.cache = {}
        self.factorials = [1] * init_size
        self.max_n = 1
        self.cur_array_size = init_size

    def __call__(self, matches, size, match_prob=0.25, mismatch_prob=0.75):
        """Computes the random-match probability for a given sequence size and
        number of matches.

        Args:
            match_prob: Probability of two random bases matching.
            mismatch_prob: Probability of two random bases not matcing.

        Returns:
            The probability.
        """
        # First see if we have the result in the cache
        key = (matches, size, match_prob)
        prob = self.cache.get(key, None)
        if prob:
            return prob

        # When there are no mismatches, the probability is
        # just that of observing a specific sequence of the
        # given length by chance.
        if matches == size:
            prob = match_prob ** matches

        else:
            nfac = self.factorial(size)
            prob = 0.0

            for i in range(matches, size+1):
                j = size - i
                # use integer division in the case that the numbers are too
                # large for floating point division
                try:
                    div = nfac / self.factorial(i) / self.factorial(j)
                except OverflowError:
                    div = nfac // self.factorial(i) // self.factorial(j)
                prob += (mismatch_prob ** j) * (match_prob ** i) * div

        self.cache[key] = prob
        return prob

    def factorial(self, num):
        """Returns `num`!.
        """
        if num > self.max_n:
            self._fill_upto(num)
        return self.factorials[num]

    def _fill_upto(self, num):
        if num >= self.cur_array_size:
            extension_size = num - self.cur_array_size + 1
            self.factorials += [1] * extension_size
        idx = self.max_n
        next_i = idx + 1
        while idx < num:
            self.factorials[next_i] = next_i * self.factorials[idx]
            idx = next_i
            next_i += 1
        self.max_n = idx

class Mergeable(object):
    """Base class for objects that can merge themselves with another.
    """
    def merge(self, other):
        """Merges `other` with `self` and returns the merged value.
        """
        raise NotImplementedError()

class Summarizable(object):
    """Base class for objects that can summarize themselves.
    """
    def summarize(self):
        """Returns a summary dict.
        """
        raise NotImplementedError()

class Const(Mergeable):
    """A :class:`Mergeable` that is a constant value. Merging simply checks
    that two values are identical.

    Args:
        value: The value to treat as a constant.
    """
    def __init__(self, value):
        self.value = value

    def merge(self, other):
        """Check that `self==other`.

        Raises:
            ValueError
        """
        if self != other:
            raise ValueError("{} != {}".format(self, other))
        return self

    def __eq__(self, other):
        """Returns True if `self.value==other` (or `other.value` if `other` is
        a :class:`Const`).
        """
        if isinstance(other, Const):
            other = other.value
        return self.value == other

    def __repr__(self):
        return str(self.value)

class Timestamp(object):
    """Records datetime and clock time at object creation.
    """
    def __init__(self):
        self.dtime = datetime.now()
        self.process_time = time.process_time()

    def timestamp(self):
        """Returns the unix timestamp.
        """
        return self.dtime.timestamp()

    def isoformat(self):
        """Returns the datetime in ISO format.
        """
        return self.dtime.isoformat()

    def __sub__(self, other, minval=0.01):
        """Subtract another timestamp from this one.

        Args:
            other: The other timestamp.
            minval: The minimum difference.

        Returns:
            A dict of {wallclock=<datetime_diff>, cpu=<clock_diff>}.
        """
        return dict(
            wallclock=max(minval, self.timestamp() - other.timestamp()),
            cpu=max(minval, self.process_time - other.process_time))

class Timing(Summarizable):
    """Context manager that maintains timing information using
    :class:`Timestamp`s. Maintains a start time on __enter__, and can be updated
    with the current time by the user. Does a final update on __exit__.
    """
    def __init__(self):
        self.start_time = None
        self.cur_time = None

    def __enter__(self):
        self.start_time = Timestamp()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.update()

    def update(self):
        """Set :attr:`self.cur_time` to the current time.
        """
        self.cur_time = Timestamp()

    def summarize(self):
        """Returns a summary dict
        {start=<start_time>, wallclock=<datetime_diff>, cpu=<clock_diff>}.
        """
        if not self.cur_time:
            self.update()
        assert self.start_time is not None
        summary = dict(start=self.start_time.isoformat())
        summary.update(self.cur_time - self.start_time)
        return summary

class CountingDict(dict, Mergeable, Summarizable):
    """A dictionary that always returns 0 on get of a missing key.

    Args:
        sort_by: Whether summary is sorted by key (0) or value (1).
    """
    def __init__(self, keys=None, sort_by=0, summary_type='dict'):
        super().__init__()
        self.sort_by = sort_by
        self.summary_type = summary_type
        if keys:
            for key in keys:
                self.increment(key)

    def __getitem__(self, name):
        return self.get(name, 0)

    def increment(self, key, inc=1):
        """Increment the count of `key` by `inc`.
        """
        self[key] += inc

    def merge(self, other):
        if not isinstance(other, CountingDict):
            raise ValueError(
                "Cannot merge object of type {}".format(type(other)))
        for key, value in other.items():
            self[key] += value
        return self

    def get_sorted_items(self):
        """Returns an iterable of (key, value) sorted according to this
        CountingDict's `sort_by` param.
        """
        return sorted(self.items(), key=lambda item: item[self.sort_by])

    def summarize(self):
        """Returns an OrderedDict of sorted items.
        """
        summary_func = ordered_dict if self.summary_type == 'dict' else tuple
        return summary_func(self.get_sorted_items())

class Histogram(CountingDict):
    """Counting dict that returns a summary dict that contains summary stats.
    """
    def summarize(self):
        hist = super().summarize()
        return dict(
            hist=hist,
            summary=self.get_summary_stats())

    def get_summary_stats(self):
        """Returns dict with mean, median, and modes of histogram.
        """
        values = tuple(self.keys())
        counts = tuple(self.values())
        mu0 = weighted_mean(values, counts)
        return dict(
            mean=mu0,
            stdev=weighted_stdev(values, counts, mu0),
            median=weighted_median(values, counts),
            modes=weighted_modes(values, counts))

class NestedDict(dict, Mergeable, Summarizable):
    """A dict that initalizes :class:`CountingDict`s for missing keys.

    Args:
        shape: The flattened shape: 'long' or 'wide'.
    """
    def __init__(self, shape="wide"):
        super().__init__()
        self.shape = shape

    def __getitem__(self, name):
        if name not in self:
            self[name] = CountingDict()
        return self.get(name)

    def merge(self, other):
        if not isinstance(other, NestedDict):
            raise ValueError(
                "Cannot merge object of type {}".format(type(other)))
        for key, value in other.items():
            if key in self:
                self[key].merge(value)
            else:
                self[key] = value
        return self

    def summarize(self):
        """Returns a flattened version of the nested dict.

        Returns:
            When `shape=='long'`, a list of (key1, key2, value) tuples.
            When `shape=='wide'`, a dict of
                {columns:keys2, rows: {key1, values}}, where `keys2` is the set
                of keys in the child dicts.
        """
        keys1 = sorted(self.keys())
        if self.shape == "long":
            return tuple(
                (key1, key2, value)
                for key1 in keys1
                for key2, value in self[key1].items())
        else:
            keys2 = set()
            for child in self.values():
                keys2.update(child.keys())
            keys2 = tuple(sorted(keys2))
            return dict(
                columns=keys2,
                rows=ordered_dict(
                    (key1, tuple(self[key1].get(key2, 0) for key2 in keys2))
                    for key1 in keys1))

class MergingDict(OrderedDict, Mergeable):
    """An :class:`collections.OrderedDict` that implements :class:`Mergeable`.
    """
    def merge(self, other):
        """Merge `other` with `self` using :method:`merge_dicts`.
        """
        merge_dicts(self, other)
        return self

def merge_dicts(dest, src):
    """Merge corresponding items in `src` into `dest`. Values in `src` missing
    in `dest` are simply added to `dest`. Values that appear in both `src` and
    `dest` are merged using `merge_values`.

    Args:
        dest: The dict to merge into.
        src: The dict to merge from.

    Raises:
        ValueError if a value is not one of the accepted types.
    """
    for key, v_src in src.items():
        if dest.get(key, None) is None:
            dest[key] = v_src
        elif v_src is not None:
            dest[key] = merge_values(dest[key], v_src)

def merge_values(v_dest, v_src):
    """Merge two values based on their types, as follows:

    - Mergeable: merging is done by the dest object's merge function.
    - dict: merge_dicts is called recursively.
    - Number: values are summed.
    - Iterable (non-string): First src and dest values are converted to tuples;
    they must be the same length. Then, corresponding values are handled as
    above. The value is returned as a list.
    - Otherwise: Treated as a Const (i.e. must be identical).

    Args:
        v_dest: The dest value.
        v_src: The src value.

    Returns:
        The merged value.
    """
    if isinstance(v_dest, Mergeable):
        v_dest = v_dest.merge(v_src)
    elif isinstance(v_dest, dict):
        assert isinstance(v_src, dict)
        merge_dicts(v_dest, v_src)
    elif isinstance(v_dest, str):
        assert v_dest == v_src
    elif isinstance(v_dest, Number):
        v_dest += v_src
    elif isinstance(v_dest, Iterable):
        i_dest = tuple(v_dest)
        i_src = tuple(v_src)
        if len(i_dest) == 0:
            v_dest = i_src
        elif len(i_src) > 0:
            v_dest = [merge_values(d, s) for d, s in zip(i_dest, i_src)]
    else:
        assert v_dest == v_src
    return v_dest

def ordered_dict(iterable):
    """Create an OrderedDict from an iterable of (key, value) tuples.
    """
    ordict = OrderedDict()
    for key, value in iterable:
        ordict[key] = value
    return ordict

def complement(seq):
    """Returns the complement of nucleotide sequence `seq`.
    """
    return "".join(BASE_COMPLEMENTS[base] for base in seq)

def reverse_complement(seq):
    """Returns the reverse complement of nucleotide sequence `seq`.
    """
    return "".join(BASE_COMPLEMENTS[base] for base in reversed(seq))

def sequence_complexity(seq):
    """Computes a simple measure of sequence complexity.

    Args:
        seq: The sequence to measure.

    Returns:
        Complexity, as a value [0,2], where 0 = a homopolymer and
        2 = completely random.
    """
    seq = seq.upper()
    seqlen = float(len(seq))
    term = 0
    for base in ('A','C','G','T'):
        count = seq.count(base)
        if count > 0:
            frac = count / seqlen
            term += frac * math.log(frac) / LOG2
    return -term

def qual2int(qual, base=33):
    """Convert a quality charater to a phred-scale int.

    Args:
        q: The quality value.
        base: The offset of the first quality value (Old Illumina = 64,
            new Illumina = 33).

    Returns:
        The integer quality.
    """
    return ord(qual) - base

def quals2ints(quals, base=33):
    """Convert an iterable of quality characters to phred-scale ints.

    Args:
        quals: The qualities.
        base: The offset of the first quality value (Old Illumina = 64,
            new Illumina = 33).

    Returns:
        A tuple of integer qualities.
    """
    return (ord(q) - base for q in quals)

def qual2prob(qchar):
    """Converts a quality char to a probability.
    """
    return 10 ** (-qual2int(qchar) / 10)

def enumerate_range(collection, start, end):
    """Generates an indexed series:  (0,coll[0]), (1,coll[1]) ...

    Only up to (start-end+1) items will be yielded.

    Args:
        collection: The collection to enumerate.
        start: The starting index.
        end: The ending index.

    Yields:
        (index, item) tuples.
    """
    idx = start
    itr = iter(collection)
    while idx < end:
        yield (idx, next(itr))
        idx += 1

def mean(values):
    """Computes the mean of a sequence of numeric values.

    Args:
        values: Sequence of numeric values.

    Returns:
        The mean (floating point).
    """
    if len(values) == 0:
        raise ValueError("Cannot determine the mode of an empty sequence")
    return sum(values) / len(values)

def weighted_mean(values, counts):
    """Computes the mean of a sequence of numeric values weighted by counts.

    Args:
        values: Sequence of numeric values.

    Returns:
        The weighted mean (floating point).
    """
    datalen = len(values)
    if datalen == 0:
        raise ValueError("Cannot determine the mena of an empty sequence")
    if datalen != len(counts):
        raise ValueError("'values' and 'counts' must be the same length")
    return sum(v * c for v, c in zip(values, counts)) / sum(counts)

def stdev(values, mu0=None):
    """Returns standard deviation of values having the specified mean.
    """
    datalen = len(values)
    if datalen == 0:
        raise ValueError("Cannot determine the stdev of an empty sequence")
    if datalen == 1:
        return 0
    if mu0 is None:
        mu0 = mean(values)
    return math.sqrt(sum((val - mu0) ** 2 for val in values) / len(values))

def weighted_stdev(values, counts, mu0=None):
    """Returns standard deviation of values having the specified mean weighted
    by counts.
    """
    datalen = len(values)
    if datalen == 0:
        raise ValueError("Cannot determine the stdev of an empty sequence")
    if datalen != len(counts):
        raise ValueError("'values' and 'counts' must be the same length")
    if datalen == 1:
        return 0
    if mu0 is None:
        mu0 = weighted_mean(values, counts)
    return math.sqrt(
        sum(
            ((val - mu0) ** 2) * count
            for val, count in zip(values, counts)) /
        sum(counts))

def median(values):
    """Median function borrowed from python statistics module, and sped up by
    in-place sorting of the array.

    Args:
        data: Sequence of numeric values.

    Returns:
        The median (floating point).
    """
    datalen = len(values)
    if datalen == 0:
        raise ValueError("Cannot determine the median of an empty sequence")

    values.sort()

    idx = datalen // 2
    if datalen % 2 == 1:
        return values[idx]
    else:
        return (values[idx - 1] + values[idx]) / 2

def weighted_median(values, counts):
    """Compute the median of `values` weighted by `counts`.

    Args:
        values: Sequence of unique values.
        counts: Sequence of counts, where each count is the number of times the
            value at the corresponding position appears in the sample.

    Returns:
        The weighted median.
    """
    datalen = len(values)
    if datalen == 0:
        raise ValueError("Cannot determine the median of an empty sequence")
    if datalen != len(counts):
        raise ValueError("'values' and 'counts' must be the same length")
    counts_cumsum = functools.reduce(
        lambda c, x: c + [c[-1] + x], counts, [0])[1:]
    total = counts_cumsum[-1]
    if total == 0:
        return None
    mid1 = mid2 = (total // 2) + 1
    if total % 2 == 0:
        mid1 -= 1
    val1 = val2 = None
    for i, val in enumerate(counts_cumsum):
        if val1 is None and mid1 <= val:
            val1 = values[i]
        if mid2 <= val:
            val2 = values[i]
            break
    return float(val1 + val2) / 2

def modes(values):
    """Returns a sorted sequence of the modal (i.e. most frequent) values.
    """
    datalen = len(values)
    if datalen == 0:
        raise ValueError("Cannot determine the mode of an empty sequence")
    elif datalen == 1:
        return values
    return _find_modes(CountingDict(values).items())

def weighted_modes(values, counts):
    """Returns a sorted sequence of the modal (i.e. most frequent) values
    weighted by counts.
    """
    datalen = len(values)
    if datalen == 0:
        raise ValueError("Cannot determine the mode of an empty sequence")
    if datalen != len(counts):
        raise ValueError("'values' and 'counts' must be the same length")
    if datalen == 1:
        return values
    return _find_modes(zip(values, counts))

def _find_modes(value_count_iter):
    sorted_counts = sorted(value_count_iter, key=lambda x: x[1], reverse=True)
    modal_values = [sorted_counts[0][0]]
    mode_count = sorted_counts[0][1]
    for value, count in sorted_counts[1:]:
        if count == mode_count:
            modal_values.append(value)
        else:
            break
    modal_values.sort()
    return modal_values

def truncate_string(string, max_len=100):
    """Shorten string s to at most n characters, appending "..." if necessary.
    """
    if string is None:
        return None
    if len(string) > max_len:
        string = string[:max_len-3] + '...'
    return string

def run_interruptible(func, *args, **kwargs):
    """Run a function, gracefully handling keyboard interrupts.

    Args:
        func: The function to execute.
        args, kwargs: Positional and keyword arguments to pass to `func`.

    Returns:
        A (unix-style) return code (0=normal, anything else is an error).

    Raises:

    """
    retcode = 0
    try:
        func(*args, **kwargs)
    except KeyboardInterrupt:
        logging.getLogger().error("Interrupted")
        retcode = 130
    except IOError as err:
        if err.errno == errno.EPIPE:
            retcode = 1
        else:
            raise
    except (AtroposError, EOFError):
        logging.getLogger().error("Atropos error", exc_info=True)
        retcode = 1
    except Exception: # pylint: disable=broad-except
        logging.getLogger().error("Unknown error", exc_info=True)
        retcode = 1
    return retcode



#!/usr/bin/env python
import argparse
import os
import pandas as pd
from common import fileopen
from compute_simulated_accuracy import summary_fields

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", default="-")
    parser.add_argument("-o", "--output")
    parser.add_argument("-n", "--name", default="table")
    parser.add_argument("-c", "--caption", default="")
    parser.add_argument(
        "-e", "--error-rate-file", 
        help="Table generated by adjust_error_profiles.R that maps requested"
             "to actual error rates. Required for latex output.")
    parser.add_argument(
        "-t", "--tool-name-file",
        help="File that maps profile names to display names for tools")
    parser.add_argument(
        "-f", "--formats", choices=('txt', 'tex', 'pickle'), nargs="+",
        default=['tex'])
    args = parser.parse_args()
    
    with fileopen(args.input, 'rt') as inp:
        table = pd.read_csv(inp, sep="\t", names=summary_fields)
    
    if 'txt' in args.formats:
        with fileopen(args.output + '.txt', 'wt') as out:
            table.to_csv(out, sep="\t", index=False)
        
    if 'pickle' in args.formats:
        import pickle
        with fileopen(args.output + '.pickle', 'wb') as out:
            pickle.dump(table, out, protocol=pickle.HIGHEST_PROTOCOL)
    
    if 'tex' in args.formats:
        import numpy as np
        
        adapter_cols = (
            "non-adapter reads trimmed",
            "adapter reads overtrimmed",
            "total adapter reads undertrimmed")
        adapter_pct_cols = tuple('pct' + col for col in adapter_cols)
        base_cols = (
            "overtrimmed bases",
            "undertrimmed bases")
        total_cols = (
            "reads total error",
            "bases total error")
        all_cols = adapter_cols + adapter_pct_cols + base_cols + total_cols
        new_cols = (
            "Wrongly Trimmed",
            "Over-trimmed",
            "Under-trimmed",
            "Wrongly Trimmed",
            "Over-trimmed",
            "Under-trimmed",
            "Over-trimmed",
            "Under-trimmed",
            "Total Error", 
            "Total Error"
        )
        col_map = dict(zip(all_cols, new_cols))
        
        # Since we're evaluating adatper trimming accurracy, the number of threads 
        # don't matter (there's no randomness, so the results should be the same for
        # every run), and we don't want to consider any quality trimming.
        textable = table[(table.threads==4) & (table.qcut==0)]
        # Add additional columns
        textable["total adapter reads undertrimmed"] = (
            textable["adapter reads untrimmed"] + 
            textable["adapter reads undertrimmed"])
        for adapter_col, adapter_pct_col in zip(adapter_cols, adapter_pct_cols):
            textable[adapter_pct_col] = textable[adapter_col] / textable['retained reads']
        textable["reads total error"] = textable.loc[:,adapter_cols].apply(sum, 1) / textable['retained reads']
        textable["bases total error"] = textable.loc[:,base_cols].apply(sum, 1) / textable['total ref bases']
        # Melt into tidy format
        textable = textable.melt(id_vars=['dataset', 'program'], value_vars=all_cols)
        # Add the "level" - reads/read pct/bases
        def to_level(var):
            if 'pct' in var:
                return 'pct'
            elif 'reads' in var:
                return 'reads'
            else:
                return 'bases'
        textable['datalevel'] = list(to_level(var) for var in textable.variable)
        # Replace the variable names with those we want in the final table
        textable = textable.replace({ 'variable' : col_map })
        # Finally, pivot the table into the grouped format we want to use in the latex template and sort
        textable = textable.pivot_table(index=['dataset', 'program'], columns=['datalevel', 'variable']).sort_index()
        # Drop the unnecessary first column level
        textable.columns = textable.columns.droplevel(0)
        
        # Replace dataset names with actual error rates
        if args.error_rate_file:
            with open(args.error_rate_file, 'rt') as inp:
                error_rate_table = pd.read_csv(inp, sep="\t")
            error_rate_table = error_rate_table.groupby('Requested').agg(np.mean)
            datasets = textable.index.levels[0]
            textable.index = textable.index.set_levels(
                datasets.map(lambda x: str(round(error_rate_table.loc[x,'Actual'], 3))).values, 
                'dataset')
        
        # Replace tool names with display versions
        if args.tool_name_file:
            with open(args.tool_name_file, 'rt') as inp:
                tool_name_table = pd.read_csv(inp, sep="\t", index_col='ProfileName')
            programs = textable.index.levels[1]
            textable.index = textable.index.set_levels(
                programs.map(lambda x: tool_name_table.loc[x, 'DisplayName']).values, 
                'program')
        
        # Now render the template
        from mako import exceptions
        from mako.template import Template
        table_template = Template(filename=os.path.join(
            os.path.dirname(__file__), "simulated_accuracy_table.tex"))
        tex_file = args.output + ".tex"
        with fileopen(tex_file, "wt") as o:
            try:
                o.write(table_template.render(
                    name=args.name, caption=args.caption, table=textable))
            except:
                print(exceptions.text_error_template().render())


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Parse job info from sacct. The output is two lines - the header row and the
job info row with the largest MaxRSS value.
"""
import argparse
from common import fileopen, parse_profile, parse_size
import csv
from subprocess import check_output

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", default="-")
    parser.add_argument("-o", "--output", default="-")
    parser.add_argument("-p", "--profile", nargs='+')
    args = parser.parse_args()
    
    profile = list(parse_profile(args.profile[0]))
    if len(args.profile) > 1:
        profile.append(args.profile[1])
                
    with fileopen(args.input, 'rt') as inp:
        lines = list(csv.reader(inp, delimiter="|"))
    
    assert len(lines) >= 2
    header = lines[0]
    maxrss_col = header.index('MaxRSS')
    maxrss = 0
    maxrow = None
    for row in lines[1:]:
        row_maxrss = parse_size(row[maxrss_col])
        if row_maxrss > maxrss:
            maxrss = row_maxrss
            maxrow = row
    
    with fileopen(args.output, 'wt') as out:
        print(
            *('prog', 'prog2', 'threads', 'dataset', 'qcut'), *header,
            sep="\t", file=out)
        print(*profile, *maxrow, sep="\t", file=out)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
import argparse
import pandas as pd
from common import *

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", default="-")
    parser.add_argument("-o", "--output")
    parser.add_argument(
        "-f", "--formats", choices=('txt', 'svg', 'pickle'), nargs="+",
        default=['svg'])
    parser.add_argument(
        "-t", "--tool-name-file",
        help="File that maps profile names to display names for tools")
    parser.add_argument("--exclude-discarded", action="store_true", default=False)
    args = parser.parse_args()
    
    with fileopen(args.input, 'rt') as inp:
        table = pd.read_csv(inp, sep="\t")
    
    if 'txt' in args.formats:
        with fileopen(args.output + '.txt', 'wt') as out:
            table.to_csv(out, sep="\t", index=False)
        
    if 'pickle' in args.formats:
        import pickle
        with fileopen(args.output + '.pickle', 'wb') as out:
            pickle.dump(table, out, protocol=pickle.HIGHEST_PROTOCOL)
    
    if 'svg' in args.formats:
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import seaborn as sb
        import numpy as np
        sb.set(style="whitegrid")
        
        # Since neither threads nor compression change the outcome of trimming,
        # we don't need to worry about stratifying by those metric.
        table = table[
            (table.prog == 'untrimmed') | (
                (table.threads==4) & 
                (table.prog2 != 'atropos (insert + writer)') &
                (table.prog2 != 'atropos (insert + nowriter)'))]
        table = table.drop(['prog2', 'threads', 'dataset'], 1)
        
        # Extract MAPQ values
        def table_to_quals(q):
            quals = table[table.qcut.isin([None, q]) | (table.prog == 'untrimmed')].pivot(
                index='read_idx', columns='prog', values='read1_quality'
            ).append(table[table.qcut.isin([None, q]) | (table.prog == 'untrimmed')].pivot(
                index='read_idx', columns='prog', values='read2_quality'))

            # Optionally remove rows where the read was discarded by any program
            if args.exclude_discarded:
                quals = quals[quals.apply(lambda x: all(x >= 0), 1)]

            # Create a tidy table of the number of reads above each MAPQ threshold
            # for each program, along with the delta versus untrimmed
            max_mapq = quals.apply(max, 1)
            mapq_thresholds = [1] + list(range(5, 60, 5))
            progs = quals.columns.tolist()
            above_thresholds = pd.DataFrame.from_records([
                [mapq, int(q), sum(max_mapq >= mapq)] + quals.apply(lambda x: sum(x >= mapq), 0).tolist()
                for mapq in mapq_thresholds
            ], columns=['MAPQ', 'Q', 'MaxAboveThreshold'] + progs).melt(
                id_vars=['MAPQ', 'Q', 'MaxAboveThreshold'], var_name="Program", 
                value_name="AboveThreshold")
            above_thresholds['Delta'] = np.NaN
            for mapq in mapq_thresholds:
                for p in set(progs) - set(['untrimmed']):
                    prog_above_threshold = above_thresholds.loc[
                        (above_thresholds.MAPQ == mapq) & (above_thresholds.Program == p),
                        'AboveThreshold'
                    ].values
                    untrimmed_above_threshold = above_thresholds.loc[
                        (above_thresholds.MAPQ == mapq) & (above_thresholds.Program == 'untrimmed'),
                        'AboveThreshold'
                    ].values
                    if prog_above_threshold and untrimmed_above_threshold:
                        diff = prog_above_threshold - untrimmed_above_threshold
                    elif prog_above_threshold:
                        diff = prog_above_threshold
                    elif untrimmed_above_threshold:
                        diff = -untrimmed_above_threshold
                    else:
                        diff = 0
                    above_thresholds.loc[
                        (above_thresholds.MAPQ == mapq) & (above_thresholds.Program == p),
                        'Delta'
                    ] = diff
            return above_thresholds
        
        above_thresholds = pd.concat([table_to_quals(q) for q in (0, 20)])
        
        # Replace tool names with display versions
        if args.tool_name_file:
            with open(args.tool_name_file, 'rt') as inp:
                tool_name_table = pd.read_csv(inp, sep="\t", index_col='ProfileName')
            above_thresholds.Program = above_thresholds.Program.map(
                lambda x: tool_name_table.loc[x, 'DisplayName']).values
        
        plot = sb.factorplot(
            x='MAPQ', y='Delta', hue='Program', col='Q',
            data=above_thresholds[above_thresholds.Program != 'untrimmed'])
        plot.set_xlabels('Mapping Quality Score (MAPQ) Cutoff')
        plot.set_ylabels('Difference versus Untrimmed Reads')
        svg_file = args.output + ".svg"
        plot.savefig(svg_file)

if __name__ == "__main__":
    main()

# Original R code:
# process.wgbs <- function(tab, ) {
#     progs <- c('untrimmed', sort(setdiff(unique(tab$prog), 'untrimmed')))
#     N <- max(tab$read_idx)
#     num.progs <- length(progs)
#     for (i in c(4:10, ifelse(has.regions, 20, 19))) {
#         tab[,i] <- ifelse(tab[,i]=='True', TRUE, FALSE)
#     }
#     sample_tabs <- lapply(1:num.progs, function(i) tab[seq(i,nrow(tab),num.progs),])
#     names(sample_tabs) <- unlist(lapply(sample_tabs, function(x) x[1,1]))
#     quals <- rbind(
#         data.frame(read_id=1:N, read=1, do.call(cbind, lapply(sample_tabs, function(x) x$read1_quality))),
#         data.frame(read_id=1:N, read=2, do.call(cbind, lapply(sample_tabs, function(x) x$read2_quality)))
#     )
#     if (has.regions) {
#         in_region <- rbind(
#             data.frame(read_id=1:N, read=1, do.call(cbind, lapply(sample_tabs, function(x) x$read1_in_region))),
#             data.frame(read_id=1:N, read=2, do.call(cbind, lapply(sample_tabs, function(x) x$read2_in_region)))
#         )
#     }
#     if (exclude.discarded) {
#         w <- apply(quals[,3:ncol(quals)], 1, function(x) any(x==-1))
#         if (has.regions) {
#             w <- w | apply(in_region[,3:ncol(in_region)], 1, function(x) any(x==-1))
#             in_region <- in_region[!w,]
#         }
#         quals <- quals[!w,]
#     }
#     quals$maxq <- apply(quals[,3:ncol(quals)], 1, max)
#     th_values <- c(1,seq(5,60,5))
#     pts <- melt(as.data.frame(t(sapply(th_values, function(th) {
#         c(
#             th=th,
#             x=sum(quals$maxq >= th),
#             sapply(progs, function(p) {
#                 sum(quals[,p] >= th)
#             })
#         )
#     }))), id.vars=c('th', 'x'), measure.vars=progs, variable.name='prog', value.name='y')
#     pts$delta <- NA
#     for (th in th_values) {
#         for (prog in progs) {
#             pts[pts$prog == prog & pts$th == th, 'delta'] <- pts[pts$prog == prog & pts$th == th, 'y'] - pts[pts$prog == 'untrimmed' & pts$th == th, 'y']
#         }
#     }
#     retval<-list(progs=progs, quals=quals, pts=pts)
#     if (has.regions) {
#         in_region$maxi <- apply(in_region[,3:ncol(in_region)], 1, max)
#         pts2 <- melt(as.data.frame(t(sapply(th_values, function(th) {
#             c(
#                 th=th,
#                 x=sum(in_region[quals$maxq >= th, 'max']==1),
#                 sapply(progs, function(p) {
#                     sum(in_region[quals[,p] >= th, p]==1)
#                 })
#             )
#         }))), id.vars=c('th', 'x'), measure.vars=progs, variable.name='prog', value.name='y')
#         pts2$delta <- NA
#         for (th in th_values) {
#             for (prog in progs) {
#                 pts2[pts2$prog == prog & pts2$th == th, 'delta'] <- pts2[pts2$prog == prog & pts2$th == th, 'y'] - pts2[pts2$prog == 'untrimmed' & pts2$th == th, 'y']
#             }
#         }
#         retval <- c(retval, list(reg=in_region, pts2=pts2))
#     }
#     retval
# }
# New plot function
# plot.wgbs.data <- function(data, q.labels=NULL) {
#     pts <- data$pts
#     progs <- data$progs
#     pts$prog <- as.character(pts$prog)
#     colnames(pts) <- c("MAPQ", "x", "Q", "y", "Delta")
#     if (!is.null(q.labels)) {
#         for (i in 1:length(q.labels)) {
#             pts[pts$Q == progs[i], 'Q'] <- names(q.labels)[i]
#         }
#         n <- names(sort(q.labels))
#     }
#     else {
#         n <- sort(unique(pts$Q))
#     }
#     pts$Q <- factor(pts$Q, levels=n)
#     pts <- pts[order(pts$Q),]
#     ggplot(pts[pts$Q != 'untrimmed',], aes(x=MAPQ, y=Delta, colour=Q)) +
#         geom_line() + geom_point() +
#         labs(x="Mapping Quality Score (MAPQ) Cutoff", y="Difference versus Untrimmed Reads")
# }
# Old plot function
# plot.data <- function(data, prog.labels) {
#     pts <- data$pts
#     progs <- data$progs
#     pts$prog <- as.character(pts$prog)
#     colnames(pts) <- c("MAPQ", "x", "Program", "y", "Delta")
#     for (i in 1:length(prog.labels)) {
#         pts[pts$Program == progs[i], 'Program'] <- prog.labels[i]
#     }
#     pts$Q <- 0
#     pts[grep(pts$Program, pattern = 'Q20'), 'Q'] <- 20
#     pts$Q <- factor(pts$Q)
#     ggplot(pts[pts$Program != 'Untrimmed',], aes(x=MAPQ, y=Delta, colour=Program, shape=Q)) +
#         geom_line() + geom_point() +
#         labs(x="Mapping Quality Score (MAPQ) Cutoff", y="Difference versus Untrimmed Reads")
# }



"""
Build atropos.

Cython is run when
* no pre-generated C sources are found,
* or the pre-generated C sources are out of date,
* or when --cython is given on the command line.
"""
import codecs
import os.path
import sys

from setuptools import setup, Extension, find_packages
from distutils.version import LooseVersion
from distutils.command.sdist import sdist as _sdist
from distutils.command.build_ext import build_ext as _build_ext

import versioneer

MIN_CYTHON_VERSION = "0.25.2"


if sys.version_info < (3, 3):
    sys.stdout.write("At least Python 3.3 is required.\n")
    sys.exit(1)


def out_of_date(_extensions):
    """
    Check whether any pyx source is newer than the corresponding generated
    C source or whether any C source is missing.
    """
    for extension in _extensions:
        for pyx in extension.sources:
            path, ext = os.path.splitext(pyx)
            if ext not in (".pyx", ".py"):
                continue
            if extension.language == "c++":
                csource = path + ".cpp"
            else:
                csource = path + ".c"
            # When comparing modification times, allow five seconds slack:
            # If the installation is being run from pip, modification
            # times are not preserved and therefore depends on the order in
            # which files were unpacked.
            if not os.path.exists(csource) or (
                os.path.getmtime(pyx) > os.path.getmtime(csource) + 5
            ):
                return True
    return False


def no_cythonize(_extensions, **_ignore):
    """
    Change file extensions from .pyx to .c or .cpp.

    Copied from Cython documentation
    """
    for extension in _extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in (".pyx", ".py"):
                if extension.language == "c++":
                    ext = ".cpp"
                else:
                    ext = ".c"
                sfile = path + ext
            sources.append(sfile)
        extension.sources[:] = sources


def check_cython_version():
    """Exit if Cython was not found or is too old"""
    try:
        from Cython import __version__ as cyversion
    except ImportError:
        sys.stdout.write(
            "ERROR: Cython is not installed. Install at least Cython version " +
            str(MIN_CYTHON_VERSION) + " to continue.\n")
        sys.exit(1)
    if LooseVersion(cyversion) < LooseVersion(MIN_CYTHON_VERSION):
        sys.stdout.write(
            "ERROR: Your Cython is at version '{}' but at least version '{}' "
            "is required.\n".format(cyversion, MIN_CYTHON_VERSION))
        sys.exit(1)


extensions = [
    Extension("atropos.align._align", sources=["atropos/align/_align.pyx"]),
    Extension(
        "atropos.commands.trim._qualtrim",
        sources=["atropos/commands/trim/_qualtrim.pyx"]
    ),
    Extension("atropos.io._seqio", sources=["atropos/io/_seqio.pyx"]),
]

cmdclass = versioneer.get_cmdclass()
versioneer_build_ext = cmdclass.get("build_ext", _build_ext)
versioneer_sdist = cmdclass.get("sdist", _sdist)


class BuildExt(versioneer_build_ext):
    def run(self):
        # If we encounter a PKG-INFO file, then this is likely a .tar.gz/.zip
        # file retrieved from PyPI that already includes the pre-cythonized
        # extension modules, and then we do not need to run cythonize().
        if os.path.exists("PKG-INFO"):
            no_cythonize(extensions)
        else:
            # Otherwise, this is a "developer copy" of the code, and then the
            # only sensible thing is to require Cython to be installed.
            check_cython_version()
            from Cython.Build import cythonize
            self.extensions = cythonize(self.extensions)
        _build_ext.run(self)


cmdclass["build_ext"] = BuildExt


class SDist(versioneer_sdist):
    def run(self):
        # Make sure the compiled Cython files in the distribution are up-to-date
        from Cython.Build import cythonize
        check_cython_version()
        cythonize(extensions)
        versioneer_sdist.run(self)


cmdclass["sdist"] = SDist


setup(
    name="atropos",
    version=versioneer.get_version(),
    cmdclass=cmdclass,
    author="John Didion",
    author_email="john.didion@nih.gov",
    url="https://atropos.readthedocs.org/",
    description="trim adapters from high-throughput sequencing reads",
    long_description=codecs.open(
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "README.md"
        ),
        "rb",
        "utf-8"
    ).read(),
    long_description_content_type="text/markdown",
    license="MIT",
    ext_modules=extensions,
    packages=find_packages(),
    scripts=["bin/atropos"],
    package_data={
        "atropos": [
            "adapters/*.fa",
            "commands/**/templates/*"
        ]
    },
    install_requires=["cython>={}".format(MIN_CYTHON_VERSION)],
    tests_require=["pytest"],  # , "jinja2", "pysam"],
    extras_require={
        "progressbar": ["progressbar2"],
        "tqdm": ["tqdm"],
        "khmer": ["khmer"],
        "pysam": ["pysam"],
        "jinja": ["jinja2"],
        "sra": ["srastream>=0.1.3"]
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: MIT License",
        "License :: Public Domain",
        "Natural Language :: English",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6"
    ]
)



# coding: utf-8
"""
Alignment module.
"""
from collections import namedtuple
from atropos.align._align import Aligner, MultiAligner, compare_prefixes, locate
from atropos.util import RandomMatchProbability, reverse_complement

# flags for global alignment

# The interpretation of the first flag is:
# An initial portion of seq1 may be skipped at no cost.
# This is equivalent to saying that in the alignment,
# gaps in the beginning of seq2 are free.
#
# The other flags have an equivalent meaning.
START_WITHIN_SEQ1 = 1
START_WITHIN_SEQ2 = 2
STOP_WITHIN_SEQ1 = 4
STOP_WITHIN_SEQ2 = 8

# Use this to get regular semiglobal alignment
# (all gaps in the beginning or end are free)
SEMIGLOBAL = (
    START_WITHIN_SEQ1 | START_WITHIN_SEQ2 |
    STOP_WITHIN_SEQ1 | STOP_WITHIN_SEQ2)

def compare_suffixes(
        suffix_ref, suffix_query, wildcard_ref=False, wildcard_query=False):
    """Find out whether one string is the suffix of the other one, allowing
    mismatches. Used to find an anchored 3' adapter when no indels are allowed.

    Args:
        suffix_ref, suffix_query: The suffices to compare.
        wildcard_ref, wildcard_query: Whether wildcards are valid in either of
            the suffices.
    """
    suffix_ref = suffix_ref[::-1]
    suffix_query = suffix_query[::-1]
    _, length, _, _, matches, errors = compare_prefixes(
        suffix_ref, suffix_query, wildcard_ref, wildcard_query)
    return (
        len(suffix_ref) - length, len(suffix_ref), len(suffix_query) - length,
        len(suffix_query), matches, errors)

# Common match-result object returned by aligners

# TODO creating instances of this class is relatively slow and responsible for
# quite some runtime.

class Match(object):
    """An alignment match.

    Args:
        astart: Starting position of the match within the adapter.
        astop: Ending position of the match within the adapter.
        rstart: Starting position of the match within the read.
        rstop: Ending position of the match within the read.
        matches: Number of matching bases.
        errors: Number of mismatching bases.
        front: Whether the match is to the front of the read.
        adapter: The :class:`Adapter`.
        read: The :class:`Sequence`.
        length: The length of the match.
    """
    __slots__ = [
        'astart', 'astop', 'rstart', 'rstop', 'matches', 'errors', 'front',
        'adapter', 'read', 'length']

    def __init__(
            self, astart, astop, rstart, rstop, matches, errors,
            front=None, adapter=None, read=None):
        self.astart = astart
        self.astop = astop
        self.rstart = rstart
        self.rstop = rstop
        self.matches = matches
        self.errors = errors
        self.front = self._guess_is_front() if front is None else front
        self.adapter = adapter
        self.read = read
        # Number of aligned characters in the adapter. If there are indels,
        # this may be different from the number of characters in the read.
        self.length = self.astop - self.astart
        if self.length <= 0:
            raise ValueError('Match length must be >= 0')
        if self.length - self.errors <= 0:
            raise ValueError('A Match requires at least one matching position.')
        # TODO: this assertion may not always hold now that we use both error
        # rates and probabilities
        #if self.adapter:
        #    assert self.errors / self.length <= self.adapter.max_error_rate

    def __repr__(self):
        return (
            'Match(astart={0}, astop={1}, rstart={2}, rstop={3}, matches={4}, '
            'errors={5})').format(
                self.astart, self.astop, self.rstart, self.rstop, self.matches,
                self.errors)

    def copy(self):
        """Create a copy of this Match.
        """
        return Match(
            self.astart, self.astop, self.rstart, self.rstop, self.matches,
            self.errors, self.front, self.adapter, self.read)

    def _guess_is_front(self):
        """Return whether this is guessed to be a front adapter.

        The match is assumed to be a front adapter when the first base of
        the read is involved in the alignment to the adapter.
        """
        return self.rstart == 0

    def wildcards(self, wildcard_char='N'):
        """Return a string that contains, for each wildcard character,
        the character that it matches. For example, if the adapter
        ATNGNA matches ATCGTA, then the string 'CT' is returned.

        If there are indels, this is not reliable as the full alignment
        is not available.
        """
        wildcards = [
            self.read.sequence[self.rstart + i]
            for i in range(self.length)
            if (self.adapter.sequence[self.astart + i] == wildcard_char and
                self.rstart + i < len(self.read.sequence))
        ]
        return ''.join(wildcards)

    def rest(self):
        """Returns the part of the read before this match if this is a
        'front' (5') adapter, or the part after the match if this is not
        a 'front' adapter (3'). This can be an empty string.
        """
        if self.front:
            return self.read.sequence[:self.rstart]
        else:
            return self.read.sequence[self.rstop:]

    # TODO: write test
    def get_info_record(self):
        """Returns a :class:`MatchInfo`, which contains information about the
        match to write into the info file.
        """
        seq = self.read.sequence
        qualities = self.read.qualities
        if qualities is None:
            qualities = ''
        rsize = rsize_total = self.rstop - self.rstart
        if self.front and self.rstart > 0:
            rsize_total = self.rstop
        elif not self.front and self.rstop < len(seq):
            rsize_total = len(seq) - self.rstart
        return MatchInfo(
            self.read.name,
            self.errors,
            self.rstart,
            self.rstop,
            seq[0:self.rstart],
            seq[self.rstart:self.rstop],
            seq[self.rstop:],
            self.adapter.name,
            qualities[0:self.rstart],
            qualities[self.rstart:self.rstop],
            qualities[self.rstop:],
            self.front,
            self.astop - self.astart,
            rsize, rsize_total)

MatchInfo = namedtuple("MatchInfo", (
    "read_name", "errors", "rstart", "rstop", "seq_before", "seq_adapter",
    "seq_after", "adapter_name", "qual_before", "qual_adapter", "qual_after",
    "is_front", "asize", "rsize_adapter", "rsize_total"))


class InsertAligner(object):
    """Implementation of an insert matching algorithm.

    If the inserts align, the overhangs are searched for the adapter sequences.
    Otherwise, each read is search for its adapter separately.

    This only works with paired-end reads with 3' adapters.

    Args:
        adapter1, adapter2: read1, read2 adapters.
        match_probability: Callable that calculates random match probability
            given arguments (num_matches, match_len).
        insert_max_rmp: Max random match probability for the insert match.
        adapter_max_rmp: Max random match probability for the adapter match.
        min_insert_overlap: Minimum number of bases the inserts must overlap
            to be considered matching.
        max_insert_mismatch_frac: Maximum fraction of mismatching bases between
            the inserts to be considered matching.
        min_adapter_overlap: Minimum number of bases the adapter must overlap
            the read to be considered matching.
        max_adapter_mismatch_frac: Maximum fraction of mismatching bases between
            the adapter and the read to be considered matching.
        adapter_check_cutoff: Threshold number of matching bases required before
            adapter matches are checked against random match probability.
        base_probs: Dict of (match_prob, mismatch_prob), which are the
            probabilities passed to
            :method:`atropos.util.RandomMatchProbability.__call__()`.
    """
    def __init__(
            self, adapter1, adapter2,
            match_probability=RandomMatchProbability(),
            insert_max_rmp=1E-6, adapter_max_rmp=0.001,
            min_insert_overlap=1, max_insert_mismatch_frac=0.2,
            min_adapter_overlap=1, max_adapter_mismatch_frac=0.2,
            adapter_check_cutoff=9, base_probs=None,
            adapter_wildcards=True, read_wildcards=False):
        self.adapter1 = adapter1
        self.adapter1_len = len(adapter1)
        self.adapter2 = adapter2
        self.adapter2_len = len(adapter2)
        self.match_probability = match_probability
        self.insert_max_rmp = insert_max_rmp
        self.adapter_max_rmp = adapter_max_rmp
        self.min_insert_overlap = min_insert_overlap
        self.max_insert_mismatch_frac = float(max_insert_mismatch_frac)
        self.min_adapter_overlap = min_adapter_overlap
        self.max_adapter_mismatch_frac = float(max_adapter_mismatch_frac)
        self.adapter_check_cutoff = adapter_check_cutoff
        self.base_probs = base_probs or dict(
            match_prob=0.25, mismatch_prob=0.75)
        self.adapter_wildcards = adapter_wildcards
        self.read_wildcards = read_wildcards
        self.aligner = MultiAligner(
            max_insert_mismatch_frac,
            START_WITHIN_SEQ1 | STOP_WITHIN_SEQ2,
            min_insert_overlap)

    # def __repr__(self) -> str:
    #     return f"InsertAligner<adapter1={self.adapter1}, adapter2={self.adapter2}, " \
    #            f"match_probability={self.match_probability}, " \
    #            f"insert_max_rmp={self.insert_max_rmp}, " \
    #            f"adapter_max_rmp={self.adapter_max_rmp}, " \
    #            f"min_insert_overlap={self.min_insert_overlap}, " \
    #            f"max_insert_mismatch_frac={self.max_insert_mismatch_frac}, " \
    #            f"min_adapter_overlap={self.min_adapter_overlap}, " \
    #            f"max_adapter_mismatch_frac={self.max_adapter_mismatch_frac}, " \
    #            f"adapter_check_cutoff={self.adapter_check_cutoff}, " \
    #            f"base_probs={self.base_probs}, " \
    #            f"adapter_wildcards={self.adapter_wildcards}, " \
    #            f"read_wildcards={self.read_wildcards}, " \
    #            f"aligner={self.aligner}>"

    def match_insert(self, seq1, seq2):
        """Use cutadapt aligner for insert and adapter matching.

        Args:
            seq1, seq2: Sequences to match.

        Returns:
            A :class:`Match` object, or None if there is no match.
        """
        seq_len1 = len(seq1)
        seq_len2 = len(seq2)
        seq_len = min(seq_len1, seq_len2)
        if seq_len1 > seq_len2:
            seq1 = seq1[:seq_len2]
        elif seq_len2 > seq_len1:
            seq2 = seq2[:seq_len1]

        seq2_rc = reverse_complement(seq2)

        def _match(_insert_match, _offset, _insert_match_size, _):
            if _offset < self.min_adapter_overlap:
                # The reads are mostly overlapping, to the point where
                # there's not enough overhang to do a confident adapter
                # match. We return just the insert match to signal that
                # error correction can be done even though no adapter
                # trimming is required.
                return (_insert_match, None, None)

            # TODO: this is very sensitive to the exact correct choice of
            # adapter. For example, if you specifiy GATCGGAA... and the correct
            # adapter is AGATCGGAA..., the prefixes will not match exactly and
            # the alignment will fail. We need to use a comparison that is a bit
            # more forgiving.

            def _adapter_match(insert_seq, adapter_seq, adapter_len):
                amatch = compare_prefixes(
                    insert_seq[_insert_match_size:], adapter_seq,
                    wildcard_ref=self.adapter_wildcards,
                    wildcard_query=self.read_wildcards)
                alen = min(_offset, adapter_len)
                return amatch, alen, round(alen * self.max_adapter_mismatch_frac)

            a1_match, a1_length, a1_max_mismatches = _adapter_match(
                seq1, self.adapter1, self.adapter1_len)
            a2_match, a2_length, a2_max_mismatches = _adapter_match(
                seq2, self.adapter2, self.adapter2_len)

            if (
                    a1_match[5] > a1_max_mismatches and
                    a2_match[5] > a2_max_mismatches):
                return None

            if min(a1_length, a2_length) > self.adapter_check_cutoff:
                a1_prob = self.match_probability(a1_match[4], a1_length)
                a2_prob = self.match_probability(a2_match[4], a2_length)
                if (a1_prob * a2_prob) > self.adapter_max_rmp:
                    return None

            mismatches = min(a1_match[5], a2_match[5])

            def _create_match(alen, slen):
                alen = min(alen, slen - _insert_match_size)
                _mismatches = min(alen, mismatches)
                _matches = alen - _mismatches
                return Match(0, alen, _insert_match_size, slen, _matches, _mismatches)

            return (
                _insert_match,
                _create_match(a1_length, seq_len1),
                _create_match(a2_length, seq_len2)
            )

        # # This is the old way of doing things, where we use the built-in
        # # Aligner to do a single match.
        # aligner = Aligner(
        #     seq2_rc,
        #     self.max_insert_mismatch_frac,
        #     START_WITHIN_SEQ1 | STOP_WITHIN_SEQ2,
        #     False, False)
        # aligner.min_overlap = self.min_insert_overlap
        # aligner.indel_cost = 100000
        #
        # insert_match = aligner.locate(seq1)
        #
        # if not insert_match:
        #     return None
        #
        # offset = min(insert_match[0], seq_len - insert_match[3])
        # insert_match_size = seq_len - offset
        # prob = self.match_probability(insert_match[4], insert_match_size)
        #
        # if prob > self.insert_max_rmp:
        #     return None
        #
        # return _match(insert_match, offset, insert_match_size, prob)

        # Use an aligner that returns all matches that satisfy the
        # overlap and error rate thresholds. We sort by matches and
        # then mismatches, and then check each in turn until we find
        # one with an adapter match (if any).

        insert_matches = self.aligner.locate(seq2_rc, seq1)

        if insert_matches:
            # Filter by random-match probability
            filtered_matches = []
            for insert_match in insert_matches:
                offset = min(insert_match[0], seq_len - insert_match[3])
                insert_match_size = seq_len - offset
                prob = self.match_probability(insert_match[4], insert_match_size, **self.base_probs)
                if prob <= self.insert_max_rmp:
                    filtered_matches.append((insert_match, offset, insert_match_size, prob))

            if filtered_matches:
                if len(filtered_matches) == 1:
                    return _match(*filtered_matches[0])
                else:
                    # Test matches in order of random-match probability.
                    # TODO: compare against sorting by length (which is how
                    # SeqPurge essentially does it).
                    # filtered_matches.sort(key=lambda x: x[2], reverse=True)
                    filtered_matches.sort(key=lambda x: x[3])
                    for match_args in filtered_matches:
                        match = _match(*match_args)
                        if match:
                            return match

            return None



"""Base class and functions for implementing command-line interfaces for
Atropos subcommands.
"""
from argparse import ArgumentParser, ArgumentError, HelpFormatter
import copy
import logging
from multiprocessing import cpu_count
import operator
import os
import platform
import re
import sys
import textwrap
import urllib
from atropos import __version__
from atropos.io import STDOUT, STDERR, resolve_path, check_path, check_writeable
from atropos.io.compression import splitext_compressed
from atropos.io.seqio import SINGLE, PAIRED
from atropos.util import MAGNITUDE, ALPHABETS

class BaseCommandParser(object):
    """Base class for Atropos sub-commands.

    Subclasses must define name, description, and usage members.
    """
    preamble = "Atropos version {version}"
    usage = "atropos {command} [options]"
    description = ''
    details = ''

    def __init__(self):
        self.groups = {}
        self.create_parser()
        self.add_common_options()
        self.add_command_options()

    def parse(self, args):
        """Parse args using the conifgured ArgumentParser.

        Args:
            args: Command line arguments.

        Returns:
            A Namespace-like object.
        """
        options = self.parser.parse_args(args)
        options.orig_args = copy.copy(args)
        self.setup_logging(options)
        self.validate_common_options(options)
        self.validate_command_options(options)
        return options

    def create_parser(self):
        """Create the ArgumentParser.
        """
        format_args = dict(
            name=self.name,
            version=__version__)
        self.parser = ArgumentParser(
            prog="atropos {}".format(self.name),
            usage=self.usage.format(**format_args),
            description=self.get_description(**format_args),
            formatter_class=ParagraphHelpFormatter)

    def get_description(self, **kwargs):
        description = "{}\n\n{}\n\n{}".format(*(
            part.strip()
            for part in (self.preamble, self.description, self.details)))
        return description.format(**kwargs)

    def add_group(
            self, name, title=None, description=None, mutex=False,
            required=False):
        """Add a group to the parser. The group will be stored under `name`
        and can later be retrieved via `get_group`.
        """
        if name in self.groups:
            raise ValueError("Group already exists: {}".format(name))
        if mutex:
            group = self.parser.add_mutually_exclusive_group(required)
        else:
            group = self.parser.add_argument_group(title or name, description)
        self.groups[name] = group
        return group

    def get_group(self, name):
        """If a group has already been created with `name`, return the group,
        otherwise create a new group with that name.
        """
        if name in self.groups:
            return self.groups[name]
        else:
            return self.add_group(name)

    def add_common_options(self):
        """Add common arguments to the parser.
        """
        self.parser.set_defaults(
            orig_args=None,
            paired=False,
            default_outfile=STDOUT,
            report_file=None,
            report_formats=None,
            batch_size=1000,
            counter_magnitude="M",
            sra_reader=None)
        self.parser.add_argument(
            "--debug",
            action='store_true', default=False,
            help="Print debugging information. (no)")
        self.parser.add_argument(
            "--progress",
            choices=('bar', 'msg'), default=None,
            help="Show progress. bar = show progress bar; msg = show a status "
                 "message. (no)")
        self.parser.add_argument(
            "--quiet",
            action='store_true', default=False,
            help="Print only error messages. (no)")
        self.parser.add_argument(
            "--log-level",
            choices=('DEBUG', 'INFO', 'WARN', 'ERROR'), default=None,
            help="Logging level. (ERROR when --quiet else INFO)")
        self.parser.add_argument(
            "--log-file",
            type=writeable_file, default=None, metavar="FILE",
            help="File to write logging info. (stdout)")
        self.parser.add_argument(
            '--version', action='version', version=__version__,
            help="Show version information and exit.")

        group = self.add_group("Input")
        group.add_argument(
            "-pe1",
            "--input1",
            type=readable_file, default=None, metavar="FILE1",
            help="The first input file.")
        group.add_argument(
            "-pe2",
            "--input2",
            type=readable_file, default=None, metavar="FILE2",
            help="The second input file.")
        group.add_argument(
            "-l",
            "--interleaved-input",
            type=readable_file, default=None, metavar="FILE",
            help="Interleaved input file.")
        group.add_argument(
            "-se",
            "--single-input",
            type=readable_file, default=None, metavar="FILE",
            help="A single-end read file.")
        group.add_argument(
            "--single-input-read",
            type=int, dest='input_read', choices=(1, 2), default=None,
            help="When treating an interleaved FASTQ or paired-end SAM/BAM "
                 "file as single-end, this option specifies which of the two "
                 "reads to process. (both reads used)")
        group.add_argument(
            "-sq",
            "--single-quals",
            type=readable_file, default=None, metavar="FILE",
            help="A single-end qual file.")
        group.add_argument(
            "-sra",
            "--sra-accession",
            default=None, metavar="ACCN",
            help="Accession to stream from SRA (requires optional NGS "
                 "dependency to be installed).")
        group.add_argument(
            "-f",
            "--format",
            choices=('fasta','fastq','sra-fastq','sam','bam'), default=None,
            help="Input file format. Ignored when reading csfasta/qual files. "
                 "(auto-detect from file name extension)")
        group.add_argument(
            "-Q",
            "--quality-base",
            type=positive(), default=33,
            help="Assume that quality values in FASTQ are encoded as "
                 "ascii(quality + QUALITY_BASE). This needs to be set to 64 "
                 "for some old Illumina FASTQ files. (33)")
        group.add_argument(
            "-c",
            "--colorspace",
            action='store_true', default=False,
            help="Enable colorspace mode: Also trim the color that is adjacent "
                 "to the found adapter. (no)")
        group.add_argument(
            "--max-reads",
            type=int_or_str, default=None, metavar="N",
            help="Maximum number of reads/pairs to process (no max)")
        group.add_argument(
            "--subsample",
            type=probability, default=None, metavar="PROB",
            help="Subsample a fraction of reads. (no)")
        group.add_argument(
            "--subsample-seed",
            type=int, default=None, metavar="SEED",
            help="The seed to use for the pseudorandom number generator. Using "
                 "the same seed will result in the same subsampling of reads.")
        group.add_argument(
            "--batch-size",
            type=int_or_str, metavar="SIZE",
            help="Number of records to process in each batch. (1000)")
        group.add_argument(
            "-D",
            "--sample-id",
            default=None, metavar="ID",
            help="Optional sample ID. Added to the summary output.")
        group.add_argument(
            "--alphabet",
            default=None, metavar="NAME", choices=tuple(ALPHABETS.keys()),
            help="Specify a sequence alphabet to use for validating inputs. "
                 "Currently, only 'dna' is supported. (no validation)")

    def add_command_options(self):
        """Add command-specific options. At the very least,
        "-o, --output" is required.
        """
        raise NotImplementedError()

    def setup_logging(self, options):
        """Setup logging and print an introductory message.

        Logging setup is only done if there are not already any handlers (can
        happen when this function is being called externally such as from unit
        tests).
        """
        if not logging.root.handlers:
            level = options.log_level or ("ERROR" if options.quiet else "INFO")
            level = getattr(logging, level)
            if options.log_file is None:
                # Due to backwards compatibility, log messages are sent to
                # standard output instead of standard error if the -o option is
                # used.
                stream = sys.stdout
                if options.output in (None, STDOUT, STDERR):
                    stream = sys.stderr
                handler = logging.StreamHandler(stream)
            else:
                handler = logging.FileHandler(options.log_file)
            handler.setFormatter(logging.Formatter(
                '%(asctime)s %(levelname)s: %(message)s'))
            handler.setLevel(level)
            logging.getLogger().setLevel(level)
            logging.getLogger().addHandler(handler)

        logger = logging.getLogger()
        logger.info(
            "This is Atropos %s with Python %s", __version__,
            platform.python_version())

    def validate_common_options(self, options):
        """Validate arguments to common options.
        """
        parser = self.parser

        # Find out which 'mode' we need to use.
        # TODO: unit tests for SRA streaming
        # TODO: add srastream to pypi
        if options.sra_accession:
            if options.format not in ('fastq', 'sam', 'bam', None):
                raise ValueError(
                    "Invalid file format for SRA accession: {}".format(
                        options.format))
            options.format = 'fastq'
            logging.getLogger().debug(
                "Opening reader for SRA Accession {}".format(
                    options.sra_accession))
            try:
                from srastream import SraReader
                options.sra_reader = SraReader(
                    options.sra_accession,
                    batch_size=options.batch_size or 1000)
                options.sra_reader.start()
                options.paired = options.sra_reader.paired
            except Exception:
                logging.getLogger().exception(
                    "Error while fetching accession {} from SRA".format(
                        options.sra_accession))
                parser.error("Unable to read from accession {}".format(
                    options.sra_accession))
        elif options.single_input:
            if options.input1 or options.input2 or options.interleaved_input:
                parser.error("Cannot use -se together with -pe1, -pe2, or -l")
            options.paired = False
            options.input1 = options.single_input
            options.input2 = options.single_quals
        elif options.interleaved_input and options.input_read:
            options.input1 = options.interleaved_input
            options.paired = False
        else:
            if (not options.interleaved_input and (
                    not options.input1 or not options.input2)):
                parser.error(
                    "Both '-pe1' and '-pe2' are required for paired-end "
                    "trimming. If this is an interleaved file, use '-l' "
                    "instead.")
            options.paired = True

        if options.input_read is None:
            options.input_read = PAIRED if options.paired else SINGLE

        # Set sample ID from the input file name(s)
        if options.sample_id is None:
            if options.sra_reader:
                options.sample_id = options.sra_reader.name
            else:
                fname = os.path.basename(
                    options.input1 or options.interleaved_input)
                name = splitext_compressed(fname)[0]
                if options.input2:
                    name2 = splitext_compressed(os.path.basename(options.input2))[0]
                    if name != name2:
                        for i in range(min(len(name), len(name2))):
                            if name[i] != name2[i]:
                                name = name[:i]
                                break
                if name.endswith('.'):
                    name = name[:-1]
                options.sample_id = name

        if options.quiet:
            options.progress = None
        elif options.progress and options.output == STDERR:
            logging.getLogger().warning(
                "Progress bar may corrupt output written to STDERR")

        if options.report_file in (STDOUT, STDERR) and options.quiet:
            logging.getLogger().warning(
                "Quiet mode - report will not be written to stdout")
            options.report_file = None

    def validate_command_options(self, options):
        """Validate command-specific options.
        """
        pass

# Extensions to argparse

class ParagraphHelpFormatter(HelpFormatter):
    """HelpFormatter that wraps text in paragraphs.
    """
    def _fill_text(self, text, width, indent):
        text = re.sub('[ \t]{2,}', ' ', text)
        paragraphs = [
            textwrap.fill(
                p, width, initial_indent=indent, subsequent_indent=indent)
            for p in re.split("\n\n", text)]
        return "\n\n".join(paragraphs)

class TypeWithArgs(object): # pylint: disable=no-member
    """A data type that requires additional static arguments.
    """
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, string):
        return self._do_call(string, *self.args, **self.kwargs) or string

    def _do_call(self, string, *args, **kwargs):
        """Convert and/or validate `string`.
        """
        raise NotImplementedError()

class CompositeType(object):
    """A composite of multiple data types.
    """
    def __init__(self, *types):
        self.types = types

    def __call__(self, string):
        result = string
        for datatype in self.types:
            result = datatype(result)
        return result

class ComparisonValidator(TypeWithArgs):
    """Validator that compares an argument against an expected value.
    """
    def _do_call(self, lhs, rhs, oper, expected=True):
        assert oper(lhs, rhs) == expected, "{}({}, {}) != {}".format(
            oper, lhs, rhs, expected)

class CharList(object):
    """Parses a string into a list of characters and ensures they are all in
    the choices tuple.
    """
    def __init__(self, choices):
        self.choices = set(choices)

    def __call__(self, string):
        chars = list(string)
        assert all(char in self.choices for char in chars)
        return chars

class Delimited(TypeWithArgs):
    """Splits a string argument using a delimiter.
    """
    def _do_call(
            self, string, delim=",", data_type=None, choices=None,
            min_len=None, max_len=None):
        if isinstance(string, str):
            vals = string.split(delim) if delim else (string,)
        else:
            vals = string

        if vals[0] == "*" and choices is not None:
            vals = choices

        if data_type:
            vals = [data_type(v) for v in vals]

        if min_len and len(vals) < min_len:
            raise ArgumentError(
                self, "there must be at least {} values".format(min_len))

        if max_len and len(vals) > max_len:
            raise ArgumentError(
                self, "there can be at most {} values".format(max_len))

        return vals

ACCESS = dict(
    r=os.R_OK,
    rU=os.R_OK,
    rb=os.R_OK,
    w=os.W_OK,
    wb=os.W_OK,
    x=os.X_OK
)

class AccessiblePath(TypeWithArgs):
    """Test that a path is accessible.
    """
    def _do_call(self, path, type_, mode):
        if type_ == 'f' and path in (STDOUT, STDERR):
            return path
        if 'w' in mode:
            return check_writeable(path, type_)
        else:
            return check_path(path, type_, ACCESS[mode])

class ReadwriteableFile(object):
    """Validator for a file argument that must be both readable and writeable.
    """
    def __init__(self):
        self.read_type = AccessiblePath('f', 'r')
        self.write_type = AccessiblePath('f', 'w')

    def __call__(self, string):
        path = string
        if os.path.exists(path):
            path = self.read_type(path)
        path = self.write_type(path)
        return path

def existing_path(path):
    """Test that a path exists."""
    if path == STDOUT:
        return path
    return resolve_path(path)

readable_file = CompositeType(existing_path, AccessiblePath('f', 'r'))
"""Test that a file exists and is readable."""

writeable_file = AccessiblePath('f', 'w')
"""Test that a file 1) exists and is writelable, or 2) does not exist but
is in a writeable directory.
"""

readwriteable_file = ReadwriteableFile()
"""Test that a file is both readable and writeable."""

def readable_url(url):
    """Validator for a URL that must be readable.

    Args:
        url: The URL to validate.
    """
    parsed = urllib.parse.urlparse(url)
    scheme = parsed.scheme or 'file'
    if scheme == 'file':
        filename = readable_file(parsed.path)
        return 'file:' + filename
    else:
        return url

str_list = Delimited(data_type=str)
"""Comma-delimited list of strings."""

INT_OR_STR_RE = re.compile(r"([\d\.]+)([KkMmGg]?)")

def int_or_str(arg):
    """Similar to int(), but accepts K, M, and G abbreviations.
    """
    if arg is None or isinstance(arg, int):
        return arg
    elif isinstance(arg, str):
        match = INT_OR_STR_RE.match(arg.upper())
        num, mult = match.groups()
        if mult:
            return int(float(num) * MAGNITUDE[mult])
        else:
            return int(num)
    else:
        raise ValueError("Unsupported type {}".format(arg))

def positive(type_=int, inclusive=False):
    """Test that a number is greater than (or equal to, if ``inclusive=True``)
    zero.
    """
    oper = operator.ge if inclusive else operator.gt
    return CompositeType(type_, ComparisonValidator(0, oper))

def between(min_val=None, max_val=None, type_=int):
    """Returns a CompositeType that validates `min_val <= x <= max_val`.
    """
    return CompositeType(
        type_,
        ComparisonValidator(min_val, operator.ge),
        ComparisonValidator(max_val, operator.le))

probability = between(0, 1, float)
"""A float between 0-1 (inclusive)."""

def configure_threads(options, parser):
    """Determine the number of threads to use from the command-line options.
    Updates the value in options, and returns the number of threads.
    """
    if options.debug:
        parser.error("Cannot use debug mode with multiple threads")
    threads = options.threads
    if threads <= 0:
        threads = cpu_count()
    elif threads == 1:
        parser.error("--threads must be >= 2")
    options.threads = threads
    return threads

def parse_stat_args(args_str):
    """Parse the optional value to the '--stat' option.
    """
    args = {}
    for arg in args_str.split(';'):
        arg_parts = arg.split('=')
        if len(arg_parts) == 1:
            args[arg_parts[0]] = True
        else:
            args[arg_parts[0]] = arg_parts[1]
    return args
